
import re

from deprecated import deprecated

from pydantic import Field, model_validator
from labfreed.labfreed_infrastructure import LabFREED_BaseModel, ValidationMessage, ValidationMsgLevel
from labfreed.pac_id.pac_id import PAC_ID
from labfreed.pac_id_resolver.services import Service, ServiceGroup
from labfreed.pac_id_resolver.resolver_config_common import ( _add_msg_to_cit_entry_model, 
                                                 _validate_service_name, 
                                                 _validate_application_intent, 
                                                 _validate_service_type,
                                                 ServiceType)


@deprecated("Use ResolverConfig")
class CITEntry_v1(LabFREED_BaseModel):
    applicable_if: str = Field(..., min_length=1)
    service_name: str = Field(..., min_length=1)
    application_intent:str = Field(..., min_length=1)
    service_type:ServiceType|str
    template_url:str = Field(..., min_length=1)
        
    
    @model_validator(mode='after')
    def _validate_model(self):
        if self.applicable_if:
            conditions = self.applicable_if.split(';')
            for c in conditions:
                if '=' in c:
                    query, expected = c.split('=')
                    query = query.strip()                      
                else:
                    query = c.strip()
                    
                try:
                    # use this function to check if the pattern is valid. it returns a PatternError if not
                    _find_pattern_in_pac(query, '')
                except PatternError:
                    self._add_validation_message(
                        level=ValidationMsgLevel.ERROR,
                        source=f'Service {self.service_name}',
                        msg=f'Applicable if contains invalid pattern {query}',
                        highlight_sub=query
                    )
                except Exception:
                    pass # if no PatternError everything is fine
        return self
    
    @model_validator(mode='after')
    def _validate_service_name(self):
        msg_dict= _validate_service_name(self.service_name)
        return _add_msg_to_cit_entry_model(msg_dict, self)
    
    
    @model_validator(mode='after')
    def _validate_application_intent(self):
        msg_dict= _validate_application_intent(self.application_intent)
        return _add_msg_to_cit_entry_model(msg_dict, self)
    
    
    @model_validator(mode='after')
    def _validate_service_type(self):
        msg_dict= _validate_service_type(self.service_type)
        return _add_msg_to_cit_entry_model(msg_dict, self)
    
    
    
    

@deprecated("Use ResolverConfig")
class CIT_v1(LabFREED_BaseModel):
    origin:str = ''
    entries:list[CITEntry_v1]
    
    
    @classmethod
    def from_csv(cls, csv:str, origin=''):
        lines = csv.splitlines()
        entries = list()
        errors = list()
        for line in lines:
            if not line: # empty line
                continue
            if line.strip()[0] == '#': #comment line
                continue
            if 'Service Name' in line.strip() : #header line
                continue
            
            cols = [c.strip() for c in line.split('\t')]
            if len(cols) < 5:
                msg = ValidationMessage(
                    level=ValidationMsgLevel.ERROR,
                    source='CIT line',
                    source_id=0,
                    msg=f'Invalid line in CIT. There are {5 - len(cols)} columns missing.',
                    highlight_sub_patterns=line
                )
                errors.append(msg)
                continue
            if len(cols) > 5:
                msg = ValidationMessage(
                    level=ValidationMsgLevel.ERROR,
                    source='CIT line',
                    source_id=0,
                    msg=f'Invalid line in CIT. There are {len(cols) -5} too many columns',
                    highlight_sub_patterns=line
                )
                errors.append(msg)
                continue
            try:
                
                entry = CITEntry_v1(
                                    service_name         = cols[0],
                                    application_intent   = cols[1],
                                    service_type         = cols[2],
                                    applicable_if        = cols[3],
                                    template_url         = cols[4]
                                    )
                entries.append(entry) 
            except ValueError:
                msg = ValidationMessage(
                    level=ValidationMsgLevel.ERROR,
                    source='CIT line',
                    source_id=0,
                    msg='Invalid line in CIT.',
                    highlight_sub_patterns=line
                )
                errors.append(msg)

        cit = CIT_v1(origin=origin, entries=entries)
        if not cit.is_valid:
            errors.insert(0, 
                          ValidationMessage(
                                level=ValidationMsgLevel.WARNING,
                                source='CIT ',
                                source_id=0,
                                msg='Invalid lines in CIT. The lines were ignored. The rest of the CIT is still functional',
                                highlight_sub_patterns=''
                          )
            )
        cit._validation_messages.extend(errors)
        cit._csv_original = csv
        return cit
    
    def evaluate_pac_id(self, pac:PAC_ID):
        if type(pac) is not PAC_ID:
            raise ValueError('CIT v1 does only handle PAC-IDs. PAC-CAT it does not know what to do')
        cit_evaluated = ServiceGroup(origin=self.origin)   
        for e in self.entries:
            if e.errors():
                continue #make this stable against errors in the cit
            
            conditions = e.applicable_if.split(';')
            conditions_evaluated = list()
            for c in conditions:
                if '=' in c:
                    query, expected = c.split('=')
                    value = _find_pattern_in_pac(query.strip(), pac)
                    conditions_evaluated.append(value == expected.strip())
                else:
                    query = c.strip()
                    found = _find_pattern_in_pac(query, pac)
                    conditions_evaluated.append(found)
            is_applicable = all(conditions_evaluated)
                    
            if not is_applicable:
                continue

            url = re.sub(r"\{([^}]+)\}", lambda v: _find_pattern_in_pac(v.group(0), pac), e.template_url)
            cit_evaluated.services.append(Service(  
                                                    service_name=e.service_name,
                                                    application_intents= [ e.application_intent ],
                                                    service_type=e.service_type,
                                                    url = url
                                )
                            )            
        return cit_evaluated
    
  
    
    def __str__(self):
        if csv:=self._csv_original:
            return csv
        
        s = "# coupling information table version: 1.0\n"
        s += "Service Name\tApplication Intent\tService Type\tApplicable If\tTemplate Url\n"
        for e in self.entries:
            s += '\t'.join([e.service_name, e.application_intent, e.service_type.value, e.applicable_if, e.template_url]) + '\n'
        return s
    
    # hash and equal are only used to avoid adding the same cit multiple times. 
    # we can live with some instances, where it does not work 
    def __hash__(self):
        return self.model_dump_json().__hash__()
    
    def __eq__(self, other):
        if not isinstance(other, CIT_v1):
            return False
        return self.model_dump() == other.model_dump()



def _find_pattern_in_pac(value, pac:PAC_ID|str):
    if not isinstance(pac, str):
        pac_url =pac.to_url()
    else:
        pac_url = pac
        
    if value == '{isu}':
        return pac.issuer
    
    elif value == '{pac}':
        return pac_url.split('*')[0]
    
    elif value == '{id}':
        m = re.match(r'^HTTPS://.+?/(.+?)(\*.*)*$', pac_url)
        return m.group(1) if m else None
    
    elif m := re.match(r'\{idSeg(\d+)\}', value):
        i = int(m.group(1)) - 1 # CIT is 1 based
        seg = pac.identifier[i] if i < len(pac.identifier) else None
        if seg:
            return f"{(seg.key + ':') if seg.key else ''}{seg.value}"
        
    elif m := re.match(r'\{idVal(\w+)\}', value):
        k = m.group(1)
        seg = [s for s in pac.identifier if s.key and s.key == k]
        if seg:
            seg = seg[0]
            return seg.value   
        else:
            return None 
        
    elif value == '{ext}':
        m = re.match(r'^.*?(\*.*)*$', pac_url)
        ext_str = m.group(1) if m else None
        return m.group(1)[1:] if ext_str else None
    
    elif m := re.match(r'\{ext(\d+)\}', value):
        i = int(m.group(1)) - 1 # CIT is 1 based
        extensions = pac_url.split('*') 
        extensions.pop(0)# first element is not extension
        return extensions[i] if i < len(extensions) else None
    else:
        raise PatternError(f'{value} is not a recognized pattern for applicable if')
    
class PatternError(ValueError):
    pass
            

        
    


from __future__ import annotations

import logging
import os
from datetime import datetime
from typing import Optional, Tuple, List, Dict
from urllib.parse import urlparse, urlsplit, urlunsplit, parse_qsl, urlencode

from cachetools import TTLCache, cached

from labfreed.pac_attributes.api_data_models.response import AttributeGroup
from labfreed.pac_attributes.pythonic.py_attributes import pyAttribute, pyAttributes
from labfreed.pac_attributes.server.server import AttributeGroupDataSource
from labfreed.pac_cat.pac_cat import PAC_CAT

try:
    from openpyxl import load_workbook
except ImportError:
    raise ImportError("Please install labfreed with the [extended] extra: pip install labfreed[extended]")

# ---------------------------------------------------------------------
# Cache (shared by all instances). TTL can be overridden per instance.
# ---------------------------------------------------------------------
_cache = TTLCache(maxsize=128, ttl=0)


# ---------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------
def _is_sharepoint_url(s: str) -> bool:
    try:
        parsed = urlparse(s)
        if parsed.scheme not in {"http", "https"}:
            return False
        host = (parsed.netloc or "").lower()
        return ("sharepoint.com" in host) or ("1drv.ms" in host)
    except Exception:
        return False


def _is_local_path(s: str) -> bool:
    return os.path.exists(s) if not _is_sharepoint_url(s) else False


def _ensure_download_query(u: str) -> str:
    parts = urlsplit(u)
    q = dict(parse_qsl(parts.query, keep_blank_values=True))
    # Preserve all params (guest tokens, etc.), just force file response.
    q["download"] = "1"
    return urlunsplit((parts.scheme, parts.netloc, parts.path, urlencode(q, doseq=True), parts.fragment))


def _get_row_by_first_cell(sheet_rows: List[tuple], match_value: str, base_url: str) -> Optional[Dict[str, object]]:
    if not sheet_rows:
        return None
    headers = sheet_rows[0]
    for row in sheet_rows[1:]:
        if not row:
            continue
        first = str(row[0]).strip() if row[0] is not None else ""
        if first == match_value:
            return {
                base_url + str(headers[i]).strip(): row[i]
                for i in range(1, len(headers))
                if headers[i] is not None
            }
    return None


# ---------------------------------------------------------------------
# Base class
# ---------------------------------------------------------------------
class _BaseExcelAttributeDataSource(AttributeGroupDataSource):
    """
    Common mapping logic from Excel rows to AttributeGroup.
    Subclasses implement `_read_rows_and_last_changed()`.
    """

    def __init__(self, *, base_url: str = "", cache_duration_seconds: int = 0, uses_pac_cat_short_form:bool=True, pac_to_key=None, header_mappings=None, **kwargs):
        self._base_url = base_url
        self._uses_pac_cat_short_form = uses_pac_cat_short_form
        self._pac_to_key = pac_to_key
        self._header_mappings = header_mappings or dict()
        # allow instance-level TTL override
        try:
            _cache.ttl = int(cache_duration_seconds)
        except Exception:
            pass
        super().__init__(**kwargs)

    def is_static(self) -> bool:
        return False

    def _read_rows_and_last_changed(self) -> Tuple[List[tuple], Optional[datetime]]:
        raise NotImplementedError

    @property
    def provides_attributes(self) -> List[str]:
        rows, _ = self._read_rows_and_last_changed()
        if not rows:
            return []
        return [self._base_url + r for r in rows[0][1:]]

    def attributes(self, pac_url:str) -> Optional[AttributeGroup]:
        try:
            p = PAC_CAT.from_url(pac_url)
            pac_url = p.to_url(use_short_notation=self._uses_pac_cat_short_form, include_extensions=self._include_extensions)
            print(f'Lookup in Excel of {pac_url}')
        except:
            ... # might as well try to match the original input
            
        if f:= self._pac_to_key:
            key = f(pac_url)
        else:
            key = pac_url
            
        rows, last_changed = self._read_rows_and_last_changed()
        d = _get_row_by_first_cell(rows, key, self._base_url)
        if not d:
            return None
        attributes = [pyAttribute(key= self._header_mappings.get(k, k), value=v) for k, v in d.items() if v is not None]
        return AttributeGroup(
            group_key=self._attribute_group_key,
            attributes=pyAttributes(attributes).to_payload_attributes()
        )


# ---------------------------------------------------------------------
# Local file implementation
# ---------------------------------------------------------------------
class LocalExcelAttributeDataSource(_BaseExcelAttributeDataSource):
    def __init__(self, file_path: str, **kwargs):
        self._file_path = file_path
        super().__init__(**kwargs)

    @cached(_cache)
    def _read_rows_and_last_changed(self) -> Tuple[List[tuple], Optional[datetime]]:
        logging.info(f"Attempting to load workbook: {self._file_path!r}")

        try:
            wb = load_workbook(
                filename=self._file_path,
                read_only=True,
                data_only=True
            )
            ws = wb.active
            logging.info(f"Workbook opened successfully. Active sheet: {ws.title!r}")

            rows = list(ws.iter_rows(values_only=True))
            logging.info(f"Read {len(rows)} rows from {self._file_path!r}")

            last_changed = wb.properties.modified
            logging.info(f"Workbook 'modified' property: {last_changed}")

            wb.close()
            return rows, last_changed

        except FileNotFoundError:
            logging.error(f"Workbook not found at: {self._file_path!r}", exc_info=True)
            raise
        except PermissionError:
            logging.error(f"Permission denied when accessing: {self._file_path!r}", exc_info=True)
            raise
        except Exception as e:
            logging.exception(f"Unexpected error reading workbook {self._file_path!r}: {e}")
            raise


# # ---------------------------------------------------------------------
# # SharePoint/OneDrive (anonymous only)
# # ---------------------------------------------------------------------
# class SharePointExcelAttributeDataSource(_BaseExcelAttributeDataSource):
#     """
#     Anonymous 'anyone with the link' reader for SharePoint/OneDrive Excel.
#     No MSAL, no AAD app permissions.
#     """

#     def __init__(self, url: str, *, timeout: int = 30, **kwargs):
#         self._url = url
#         self._timeout = timeout
#         super().__init__(**kwargs)

#     @cached(_cache)
#     def _read_rows_and_last_changed(self) -> Tuple[List[tuple], Optional[datetime]]:
#         content, last_changed = self._download_bytes_anon(self._url, timeout=self._timeout)
#         with io.BytesIO(content) as fh:
#             wb = load_workbook(filename=fh, read_only=True, data_only=True)
#             ws = wb.active
#             rows = list(ws.iter_rows(values_only=True))
#             wb_last = wb.properties.modified  # often None for streamed files
#             wb.close()
#         return rows, (wb_last or last_changed)

#     def _download_bytes_anon(self, url: str, *, timeout: int) -> Tuple[bytes, Optional[datetime]]:
#         u = _ensure_download_query(url)
#         headers = {"User-Agent": "python-requests/anon-sharepoint-downloader"}
#         resp = requests.get(u, headers=headers, timeout=timeout, allow_redirects=True)
#         resp.raise_for_status()

#         last = None
#         lm_hdr = resp.headers.get("Last-Modified")
#         if lm_hdr:
#             try:
#                 last = parsedate_to_datetime(lm_hdr)
#                 if last.tzinfo is None:
#                     last = last.replace(tzinfo=timezone.utc)
#                 else:
#                     last = last.astimezone(timezone.utc)
#             except Exception:
#                 last = None
#         return resp.content, last




__all__ = [
    "LocalExcelAttributeDataSource",
]



import re
from pydantic import BaseModel, model_validator
from labfreed.well_known_keys.unece.unece_units import unece_units


class Quantity(BaseModel):
    ''' Represents a quantity'''
    value: float|int
    unit: str | None 
    '''unit. Use SI symbols. Set to None of the Quantity is dimensionless'''
    log_least_significant_digit: int|None = None
    
    
    @model_validator(mode='before')
    @classmethod
    def transform_inputs(cls, d:dict):
        if not isinstance(d, dict):
            return d
        
        # decimals_to_log_significant_digits
        if decimals:= d.pop('decimals', None):
            d['log_least_significant_digit'] = - decimals
        
        #dimensionless_unit
        unit:str= d.get('unit')
        if unit and unit in ['1', '', 'dimensionless']:
            unit = None
            d['unit'] = unit
        
        #try to coerce to ucum. catch the two most likely mistakes to use blanks for multiplication and ^ for exponents.
        if unit:
            unit = unit.replace('/ ', '/').replace(' /', '/').replace(' ', '.').replace('^', '').replace('·','.')
            d['unit'] = unit
            
        return d
    
        
    @model_validator(mode='after')
    def significat_digits_for_int(self):
        if isinstance(self.value, int):
            self.log_least_significant_digit = 0
        return self
        
    @property
    def float(self) -> float:
        ''' for clarity returns the value'''
        return self.value
    
    @classmethod
    def from_str_value(cls, value:str, unit:str|None, log_least_significant_digit=None):
        '''
        Creates a quantity from a string representing a number (e.g. -12.345E-8 ).
        It does some magic to find the least significant digit. NOTE: for numbers like 11000 it is ambiguous if the 
        trailing zeros are significant. They will be treated as significant. Use scientific notation to be specific (i.e. 11e3 if the zeros are not significant)
        '''
        if '.' not in value and 'E' not in value: 
            num_val = int(value)
        else:
            num_val = float(value)  
        
        
        if log_least_significant_digit is None:
            log_least_significant_digit = cls._find_log_significant_digits(value)
        
        q = Quantity(value = num_val, unit=unit, log_least_significant_digit=log_least_significant_digit)
        return q
    
    @classmethod
    def can_convert_to_quantity(cls, value:str):
        ''' assumes value and unit are separated by " " '''
        m = re.match(r'^\s*(?P<mantissa>-?\d+(\.\d+)?)([Ee]-?(?P<exponent>\d+))?\s+(?P<unit>\S+)\s*$', value)
        return bool(m) 
    
    @classmethod
    def from_str_with_unit(cls, value:str):
        ''' assumes value and unit are separated by " " '''
        m = re.match(r'^\s*(?P<mantissa>-?\d+(\.\d+)?)([Ee]-?(?P<exponent>\d+))?\s+(?P<unit>\S+)\s*$', value)
        if not m:
            print(f'{value} cannot be converted to Quantity')
            return None
        try:
            parts = value.strip().split(' ', 1)
            if len(parts) == 2:
                str_value = parts[0]
                unit = parts[1]
                return cls.from_str_value(str_value, unit)
        except Exception:
            return None

    @staticmethod
    def _find_log_significant_digits(value:str):
        s = value.strip()
        m = re.match(r'^(?P<mantissa>-?\d+(\.\d+)?)([Ee]-?(?P<exponent>\d+))?$', s)
        if m:
            exponent = int( m.group('exponent') or 0 )
            mantissa = m.group('mantissa')
            
        if '.' not in mantissa:
            digits_after_decimal = 0
            int_part = mantissa
            m = re.match(r'-?\d+?(?P<trailing_zeros>0*)(\.\d+)?$', int_part)
            if m:
                trailing_zeros = m.group('trailing_zeros')
                possible_non_significant_digits = len(trailing_zeros) # there is no way to know if they are really insignificant.  # noqa: F841
                log_least_significant_digit =  exponent
            
        else:
            int_part, frac_part = mantissa.rsplit('.', 1)
            digits_after_decimal = len(frac_part)
            
            log_least_significant_digit = -digits_after_decimal + exponent
            
        return log_least_significant_digit



            


    
    def value_as_str(self):
        if self.log_least_significant_digit is not None:
            log_digit = self.log_least_significant_digit
            if log_digit <= 0:
                val = f"{self.value:.{-log_digit}f}"
            else:
                factor = 10 ** log_digit
                rounded = round(self.value / factor) * factor
                val = str(int(rounded))
        else:
            val = str(self.value)
        return val
    
    def __str__(self):
        unit_symbol = self.unit
        if self.unit in [ "1", "dimensionless"] or not self.unit:
            unit_symbol = ""
        unit_symbol = unit_symbol.replace('.', '·')
        val = self.value_as_str()
        return f"{val} {unit_symbol}"
      
    def __repr__(self):
        return f'Quantity: {self.__repr__()}'
    
    
    
    
def unece_unit_code_from_quantity(q:Quantity):
        if not q.unit:
            return 'C62' # dimensionless
        by_name =   [ u['commonCode'] for u in unece_units() if u.get('name','') == q.unit] 
        by_symbol = [ u['commonCode'] for u in unece_units() if u.get('symbol','') == q.unit]
        by_code = [ u['commonCode'] for u in unece_units() if u.get('commonCode','') == q.unit]
        code = list(set(by_name) | set(by_symbol) | set(by_code))
        if len(code) != 1:
            raise ValueError(f'No UNECE unit code found for Quantity {q}' ) 
        return code[0]
    
    





import logging
import re
from types import MappingProxyType

from labfreed.labfreed_infrastructure import LabFREED_ValidationError

from labfreed.pac_id.id_segment import IDSegment
from labfreed.pac_id.extension import Extension

from labfreed.pac_cat import PAC_CAT
from labfreed.well_known_extensions import default_extension_interpreters


from typing import TYPE_CHECKING
if TYPE_CHECKING:
    # only imported during type checking
    from labfreed.pac_id import PAC_ID
    from labfreed.pac_cat import PAC_CAT

class PAC_Parser():
    '''@private
    Knows how to parse a PAC-ID. 
    From a SW engineering perspective it would be best to have no dependencies from other modules to pac_id.
    However from a Python users convenience perspective it is better to have one place where a pac url can be parsed and magically the extensions are in a meaningful type (e.g. TREX in TREX aware format) and categories are known of possible.

    >> We have given priority to convenient usage and therefore chose to have dependencies from pac_id to pac_cat and well_known_extensions
   '''
      
    @classmethod
    def from_url(cls, pac_url:str, 
                 *, 
                 extension_interpreters = 'default', 
                 try_pac_cat = True,
                 suppress_validation_errors=False
                 ) -> "PAC_ID":
        """Parses a PAC-ID with extensions

        Args:
            pac_url (str): pac id with optional extensions: e.g. HTTPS://PAC.METTORIUS.COM/-MD/BAL500/1234*N$N/ABC*SUM$TREX/A$T.A:ABC

        Raises:
            LabFREED_ValidationError: When validation fails. Note,that with suppress_errors no such error is raises

        Returns:
            PACID including extensions. If possible PAC-CAT is applied and extensions are cast to a known type, which knows how to inetrprete the data
        """
        if extension_interpreters == 'default':
            extension_interpreters = default_extension_interpreters
        
        if '*' in pac_url:
            id_str, ext_str = pac_url.split('*', 1)
        else:
            id_str = pac_url
            ext_str = ""
                    
        pac_id = cls._parse_pac_id(id_str)
        
        # try converting to PAC-CAT. This can fail, in which case a regular PAC-ID is returned
        if try_pac_cat:
            try:
                pac_cat = PAC_CAT.from_pac_id(pac_id)
                if pac_cat.categories:
                    pac_id = pac_cat
            except LabFREED_ValidationError:
                pass 
              
        extensions = cls._parse_extensions(ext_str)
        if extensions and extension_interpreters:
            for i, e in enumerate(extensions):
                if interpreter := extension_interpreters.get(e.type or ''):
                    extensions[i] = interpreter.from_extension(e)
        pac_id.extensions = extensions
            
        if not pac_id.is_valid and not suppress_validation_errors:
            logging.error(pac_id.print_validation_messages())
            raise LabFREED_ValidationError(validation_msgs = pac_id._get_nested_validation_messages())
        
        return pac_id
            
    @classmethod
    def _parse_pac_id(cls,id_str:str) -> "PAC_ID":
        # m = re.match('(HTTPS://)?(PAC.)?(?P<issuer>.+?\..+?)/(?P<identifier>.*)', id_str)
        m = re.match('(HTTPS://)?(PAC\.)?(?P<issuer>.+?)/(?P<identifier>.*)', id_str, re.IGNORECASE)
        if not m:
            raise LabFREED_ValidationError(f'{id_str} does not match the pattern expected for PAC-ID')
        d = m.groupdict()
        
        id_segments = list()
        id_segments = cls._parse_id_segments(d.get('identifier'))
        
        from labfreed.pac_id import PAC_ID
        pac = PAC_ID(issuer= d.get('issuer'),
                     identifier=id_segments
        )
        
        return pac
    
    @classmethod
    def _parse_id_segments(cls, identifier:str):
        if not identifier:
            return []    
        
        id_segments = list()  
        if len(identifier) > 0 and identifier[0] == '/':
            identifier = identifier[1:]
        for s in identifier.split('/'):
            tmp = s.split(':')
            
            if len(tmp) == 1:
                segment = IDSegment(value=tmp[0])
            elif len(tmp) == 2:
                segment = IDSegment(key=tmp[0], value=tmp[1])
            else:
                raise ValueError(f'invalid segment: {s}')
                
            id_segments.append(segment)
        return id_segments
    

    @classmethod
    def _parse_extensions(cls, extensions_str:str|None) -> list["Extension"]:    
        
        extensions = list()
        
        if not extensions_str:
            return extensions
        
        defaults =  MappingProxyType(
                                {
                                    0: { 'name': 'N', 'type': 'TEXT'},
                                    1: { 'name': 'SUM', 'type': 'TREX'}
                                }
        )
        for i, e in enumerate(extensions_str.split('*')):
            if e == '': #this will happen if first extension starts with *
                continue
            d = re.match('((?P<name>.+)\$(?P<type>.+)/)?(?P<data>.+)', e).groupdict()
            
            name = d.get('name')
            type = d.get('type') 
            data = d.get('data')
    
            if name:
                defaults = None # once a name was specified no longer assign defaults
            else:
                if defaults:
                    name = defaults.get(i).get('name')
                    type = defaults.get(i).get('type')
            
            #convert to subtype if they were given
            e = Extension.create(name=name, type=type, data=data)
            extensions.append(e)

        return extensions


from collections import Counter
from typing import Self
from pydantic import Field, field_validator

from labfreed.labfreed_infrastructure import LabFREED_BaseModel
from labfreed.trex.table_segment import _deserialize_table_segment_from_trex_segment_str
from labfreed.trex.trex_base_models import TREX_Segment
from labfreed.trex.value_segments import _deserialize_value_segment_from_trex_segment_str


class TREX(LabFREED_BaseModel):
    '''Represents a T-REX extension'''
    segments: list[TREX_Segment] = Field(default_factory=list)
       
    @classmethod
    def deserialize(cls, data) -> Self:
        segment_strings = data.split('+')
        segments = list()
        for s in segment_strings:
            # there are only two valid options. The segment is a scalar or a table. 
            # Constructors do the parsing anyways and raise exceptions if invalid data
            # try both options and then let it fail
            segment = _deserialize_table_segment_from_trex_segment_str(s)
            if not segment:
                segment = _deserialize_value_segment_from_trex_segment_str(s)
            if not segment:
                raise ValueError('TREX contains neither valid value segment nor table')
                
            segments.append(segment)
        trex = TREX(segments=segments)
        return trex
        
    
    def serialize(self):
        seg_strings = list()
        for s in self.segments:
            seg_strings.append(s.serialize())
        s_out = '+'.join(seg_strings)
        return s_out
    
       
    def get_segment(self, segment_key:str) -> TREX_Segment:
        '''Get a segment by key'''
        s = [s for s in self.segments if s.key == segment_key]
        if s:
            return s[0]
        else:
            return None
        
    
    def __str__(self):
        s = self.serialize().replace('+', '\n+').replace('::', '::\n ')
        return s
            
        
        
    @field_validator('segments')
    @classmethod
    def _validate_segments(cls, segments):
        segment_keys = [s.key for s in segments]
        duplicates = [item for item, count in Counter(segment_keys).items() if count > 1]
        if duplicates:
            raise ValueError(f"Duplicate segment keys: {','.join(duplicates)}")
        return segments
      
    
        

    



from __future__ import annotations

from dataclasses import dataclass
from datetime import UTC, datetime
from typing import Protocol, runtime_checkable
from urllib.parse import quote
import warnings

import requests

from pydantic import ValidationError

from labfreed.pac_attributes.api_data_models.request import AttributeRequestData
from labfreed.pac_attributes.api_data_models.response import AttributeGroup, AttributeResponsePayload
from labfreed.pac_attributes.client.client_attribute_group import ClientAttributeGroup
from labfreed.pac_attributes.server.server import AttributeServerRequestHandler
from labfreed.pac_id.pac_id import PAC_ID


class AuthenticationError(Exception):
    '''Server rejected authentication. '''
    pass

class AttributeClientInternalError(Exception):
    '''The error is on client side. Possibly a bug on client side'''
    pass

class AttributeServerError(Exception):
    '''The error is on server side. Needs to contact the server admin'''
    pass



@runtime_checkable
class AttributeRequestCallback(Protocol):
    def __call__(self, url: str, attribute_request_data: AttributeRequestData) -> tuple[int, str]:
        '''handle the request
        returns a tuple of HTTP status code and the body of the response or an error message'''
        ...


def http_attribute_request_default_callback_factory(session: requests.Session = None) -> AttributeRequestCallback:
    """ Returns a default implementation of AttributeRequestCallback using `requests` package.

    Args:
        session (requests.Session, optional): A requests.Session object. Pass such an object if you need control over authentication and such things. If omitted a default Session is used.

    Returns:
        AttributeRequestCallback: a callback following the AttributeRequestCallback protocol.
    """
    if session is None:
        session = requests.Session()

    def callback(url: str, attribute_request_data: AttributeRequestData) -> tuple[int, str]:
        try:
            url = url + '/' + quote(attribute_request_data.pac_id, safe='')
            params = attribute_request_data.request_params()           
            resp = session.get(url, 
                               params = params,
                               headers=attribute_request_data.language_preference_http_header(),
                               timeout=10)
            return resp.status_code, resp.text
        except requests.exceptions.RequestException as e:
            return 500, str(e)
    return callback


def local_attribute_request_callback_factory(request_handler:AttributeServerRequestHandler) -> AttributeRequestCallback:
    """ Returns a default implementation of AttributeRequestCallback using `requests` package.

    Args:
        request_handler: The request handler

    Returns:
        AttributeRequestCallback: a callback following the AttributeRequestCallback protocol.
    """

    def callback(url: str, attribute_request_data: AttributeRequestData) -> tuple[int, str]:
        try:
            resp = request_handler.handle_attribute_request(attribute_request_data)
            return 200, resp
        except requests.exceptions.RequestException as e:
            return 500, str(e)
    return callback

        
    

@dataclass 
class AttributeClient():
    """ Client handling attribute requests and caching thereof.
    """
    http_post_callback:AttributeRequestCallback
                   
    def get_attributes(self, 
                       server_url:str, 
                       pac_id:PAC_ID|str, 
                       restrict_to_attribute_groups:list[str]|None=None, 
                       language_preferences:list[str]|None=None
                       ) -> list[AttributeGroup]:
        """gets the attributes from one attribute server for one PAC-ID. Uses a cached version if possible, otherwise requests from the server again.

        Args:
            server_url (str): _description_
            pac_id (PAC_ID | str): _description_
            restrict_to_attribute_groups (list[str] | None, optional): _description_. Defaults to None.
            language_preferences (list[str] | None, optional): _description_. Defaults to None.
            force_server_request (bool, optional): _description_. Defaults to False.

        Raises:
            AuthenticationError: 
            AttributeClientInternalError: 
            AttributeServerError: 

        Returns:
            list[CacheableAttributeGroup]: attribute groups for the PAC-ID
        """
        if isinstance(pac_id, str):
            pac_id = PAC_ID.from_url(pac_id)
                
        # no valid data found in cache > request to server
        attribute_request_body = AttributeRequestData(pac_id=pac_id.to_url(), 
                                                        restrict_to_attribute_groups=restrict_to_attribute_groups,
                                                        language_preferences=language_preferences
                                )
        response_code, response_body_str = self.http_post_callback(server_url, attribute_request_body)

        if response_code == 400:
            raise AttributeClientInternalError(f"The server did not accept the request. Server message: '{response_body_str}'")
        if response_code == 401:
            raise AuthenticationError(f"Failed to authorize at the server. Server message: {response_body_str}")
        if response_code == 404:
            print(f'No atributes found for {pac_id.to_url()}')
            return []
        if response_code == 500:
            raise AttributeServerError(f"The server accepted the request, but encountered an internal error. Contact the server admin. Server message: {response_body_str}")
        try: 
            r = AttributeResponsePayload.model_validate_json(response_body_str)
        except ValidationError as e:
            print(e)
            raise AttributeServerError("The server accepted the request, and sent a reponse. However, the response is not adhering to the PAC Attributes specifications. Contact the server admin.")

        
        attribute_groups_out = []
        for ag_for_pac in r.data:
            pac_from_response = PAC_ID.from_url(ag_for_pac.pac_id)
            ags = [
                ClientAttributeGroup(
                    group_key= ag.group_key, 
                    attributes=ag.attributes, 
                    origin=server_url, 
                    language=r.language, 
                    group_label=ag.group_label
                    ) 
                for ag in ag_for_pac.attribute_groups
                ]
            
            # compare pac_id from response with pac_id we need attributes for.
            # if identical this is the part of the response we care about. other PAC-ID are just for the cache
            if pac_id.to_url() == pac_from_response.to_url():
                attribute_groups_out = ags
        
        return attribute_groups_out

            
                
    





from labfreed.pac_cat.pac_cat import PAC_CAT
from labfreed.pac_cat.predefined_categories import PredefinedCategory
from labfreed.pac_id.id_segment import IDSegment
from labfreed.pac_id.extension import Extension
from labfreed.pac_id import PAC_ID  


class PACID_Serializer():
    '''Represents a PAC-ID including it's extensions'''
  
    @classmethod
    def to_url(cls, pac:PAC_ID, use_short_notation:bool|None=None, uppercase_only=False, include_extensions:bool=True) -> str:
        """Serializes the PAC-ID including extensions.

        Args:
            use_short_notation (bool|None, optional): None (default): Preserves the identifier as is. Extensions use short form
                                                      True: forces short notation for categories and extensions
                                                      False: forces long notation for categories and extensions
            uppercase_only (bool, optional): Forces all uppercase letters (results in smaller QR)..

        Returns:
            str: Something like this HTTPS://PAC.METTORIUS.COM/-MD/BAL500/1234*N$N/ABC*SUM$TREX/A$T.A:ABC

        """
        identifier_str = cls._serialize_identifier(pac, use_short_notation=use_short_notation)
        
        if include_extensions:
            use_short_notation_for_extensions = True if use_short_notation is None else use_short_notation 
            extensions_str = cls._serialize_extensions(pac.extensions, use_short_notation=use_short_notation_for_extensions)
        else: 
            extensions_str = ""
            
        out = f"HTTPS://PAC.{pac.issuer}{identifier_str}{extensions_str}"
        
        if uppercase_only:
            out = out.upper()
        return out
    
    @classmethod
    def _serialize_identifier(cls, pac:PAC_ID|PAC_CAT, use_short_notation:None|bool=None):
        ''' Serializes the PAC-ID'''
        
        if isinstance(pac, PAC_CAT) and use_short_notation is not None:
            segments = []
            for c in pac.categories:
                segments.append(IDSegment(value=c.key))
                if isinstance(c, PredefinedCategory):
                    segments += c._get_segments(use_short_notation=use_short_notation)
                else:
                    segments += c.segments
        else:
            segments = pac.identifier
        
        identifier_str = ''
        for s in segments:
            s:IDSegment = s
            if s.key:
                identifier_str += f'/{s.key}:{s.value}'
            else:
                identifier_str += f'/{s.value}'
        return identifier_str
          

    @classmethod
    def _serialize_extensions(cls, extensions:list[Extension], use_short_notation):
        out = ''
        short_notation = use_short_notation
        for i, e in enumerate(extensions):
            
            if short_notation and i==0:
                if e.name=='N':
                    out += f'*{e.data}'
                    continue
                else: 
                    short_notation = False
            if short_notation and i==1:
                if e.name=='SUM':
                    out += f'*{e.data}'
                    continue
                else: 
                    short_notation = False
            if e.name and e.type:
                out += f'*{e.name}${e.type}/{e.data}'
            else:
                out += f'*{e.data}'
        return out
        







from .display_name_extension import DisplayNameExtension
from .trex_extension import TREX_Extension
from .default_extension_interpreters import default_extension_interpreters  # noqa: F401

__all__ = ["DisplayNameExtension", "TREX_Extension"]


from abc import ABC, abstractmethod, abstractproperty
from labfreed.pac_attributes.api_data_models.response import Attribute, AttributeGroup
from labfreed.pac_cat.pac_cat import PAC_CAT


class AttributeGroupDataSource(ABC):
    
    def __init__(self, attribute_group_key:str, include_extensions:bool=False, is_static:bool=False, ):
        self._attribute_group_key = attribute_group_key
        self._include_extensions = include_extensions
        self._is_static = is_static
       
    @property
    def is_static(self) -> bool:
        return self._is_static
    
    
    @property
    def attribute_group_key(self):
        return self._attribute_group_key
    
    @abstractproperty
    def provides_attributes(self):
        pass
    
    @abstractmethod
    def attributes(self, pac_url: str) -> AttributeGroup:
        pass
    

class Dict_DataSource(AttributeGroupDataSource):
    def __init__(self, data:dict[str, dict[str, Attribute]], uses_pac_cat_short_form=True, pac_to_key: callable = None, *args, **kwargs):
        if not all([isinstance(e, dict) for e in data.values()]):
            raise ValueError('Invalid data')
        
        self._data = data
        self.uses_pac_cat_short_form = uses_pac_cat_short_form
        self._pac_to_key = pac_to_key
        
        super().__init__(*args, **kwargs)       
        
    
    @property
    def provides_attributes(self):
        return list(set([a.key for attributes in self._data.values() for a in attributes.values()]))
    
           
    def attributes(self, pac_url: str) -> AttributeGroup:
        try:
            p = PAC_CAT.from_url(pac_url)
            pac_url = p.to_url(use_short_notation=self.uses_pac_cat_short_form, include_extensions=self._include_extensions)
        except:
            ... # might as well try to match the original input
            
        
        lookup_key = self._pac_to_key(pac_url) if self._pac_to_key else pac_url
        attributes = self._data.get(lookup_key)
        if not attributes:
            if 'default' in self._data.keys():
                attributes = self._data.get('default', None)
            else:
                return None  
        
        return AttributeGroup(group_key=self._attribute_group_key, 
                              attributes=attributes)
        
        
        

    
    







from datetime import datetime

from flask import Flask
from labfreed.pac_attributes.client.client import AttributeClient, http_attribute_request_default_callback_factory
from labfreed.pac_attributes.pythonic.attribute_server_factory import AttributeServerFactory, NoAuthRequiredAuthenticator
from labfreed.pac_attributes.pythonic.py_attributes import pyAttribute, pyAttributes, pyReference
from labfreed.pac_attributes.pythonic.py_dict_data_source import pyDict_DataSource
from labfreed.pac_attributes.server.translation_data_sources import DictTranslationDataSource, Terms, Term
from labfreed.pac_id.pac_id import PAC_ID
from labfreed.trex.pythonic.quantity import Quantity

import threading
from werkzeug.serving import make_server


MfgDate = "https://labfreed.org/terms/example/MfgDate"
MaxWeight = "https://labfreed.org/terms/example/MaxWeight"
CalWeight = "https://labfreed.org/terms/example/CalWeight"
NominalWeight = "https://labfreed.org/terms/example/NominalWeight"

data_source = pyDict_DataSource(attribute_group_key='ProductionData', 
                            data = {
                                "HTTPS://PAC.METTORIUS.COM/-MD/BAL500/12340": pyAttributes([
                                    pyAttribute(key=MfgDate, value=datetime(year=2015, month=10, day=1, hour=10, minute=12)),
                                    pyAttribute(key=MaxWeight, values=Quantity(value=100.00, unit='g', log_least_significant_digit=-2)),
                                    pyAttribute(key=CalWeight, values=pyReference('HTTPS://PAC.METTORIUS.COM/-MD/CALWEIGH/A00002'))
                                ]),
                                "HTTPS://PAC.METTORIUS.COM/-MD/BAL500/000001": pyAttributes([
                                    pyAttribute(key=MfgDate, value=datetime(year=2015, month=10, day=1, hour=10, minute=12)),
                                    pyAttribute(key=MaxWeight, values=Quantity(value=100.00, unit='g', log_least_significant_digit=-2)),
                                    pyAttribute(key=CalWeight, values=pyReference('HTTPS://PAC.METTORIUS.COM/-MD/CALWEIGH/A00002'))
                                ]),
                                "HTTPS://PAC.METTORIUS.COM/-MD/BAL500/12341": pyAttributes([
                                    pyAttribute(key=MfgDate, values=datetime(year=2015, month=10, day=1, hour=10, minute=12)),
                                    pyAttribute(key=MaxWeight, values=Quantity(value=111.00, unit='g', log_least_significant_digit=2))
                                ]),
                                "HTTPS://PAC.METTORIUS.COM/-MD/BAL500/12342": pyAttributes([
                                    pyAttribute(key=MfgDate, values=datetime(year=2015, month=10, day=1, hour=10, minute=12)),
                                    pyAttribute(key=MaxWeight, values=Quantity(value=100.00, unit='g'))
                                ]),
                                "HTTPS://PAC.METTORIUS.COM/-MD/CALWEIGH/A00002": pyAttributes([
                                    pyAttribute(key=NominalWeight, values=Quantity(value=50.00, unit='g'))
                                ])
                                }
                )


data_source_example = pyDict_DataSource(attribute_group_key='https://mettorius.com/terms/attribute_group_example', 
                            data = {
                                "HTTPS://PAC.METTORIUS.COM/-MD/BAL500/000001": pyAttributes([
                                    pyAttribute(key="https://labfreed.org/terms/example/TextAttribute", value="Bar"),
                                    pyAttribute(key="https://labfreed.org/terms/example/NumericAttribute", value="14.88 mol/L"),
                                    pyAttribute(key="https://labfreed.org/terms/example/ReferenceAttribute", value="HTTPS://PAC.METTORIUS.COM/-MD/CALWEIGH/A00002"),
                                    pyAttribute(key="https://labfreed.org/terms/example/DateTimeAttribute", value=datetime(year=2015, month=10, day=1, hour=10, minute=12)),
                                    pyAttribute(key="https://labfreed.org/terms/example/BoolAttribute", value=False),
                                    pyAttribute(key="https://labfreed.org/terms/example/ObjectAttribute", value= {'k1':1, 'k2': {'a':'bar', 'b':'foo'}, 'k3': [0,1,2]})
                                ])
                                }
                )

data_source_example_lists = pyDict_DataSource(attribute_group_key='https://mettorius.com/terms/attribute_group_example_lists', 
                            data = {
                                "HTTPS://PAC.METTORIUS.COM/-MD/BAL500/000001": pyAttributes([
                                    pyAttribute(key="https://labfreed.org/terms/example/TextAttribute", values=["Foo", 
                                                                                                                "Bar"]),
                                    pyAttribute(key="https://labfreed.org/terms/example/NumericAttribute", values=["14.88 mol/L",
                                                                                                                   "201E-3 g"]),
                                    pyAttribute(key="https://labfreed.org/terms/example/ReferenceAttribute", values=["HTTPS://PAC.METTORIUS.COM/-MD/CALWEIGH/A00002", 
                                                                                                                     "HTTPS://PAC.METTORIUS.COM/-MD/CALWEIGH/A00001"]),
                                    pyAttribute(key="https://labfreed.org/terms/example/DateTimeAttribute", values= [datetime(year=2015, month=10, day=1, hour=10, minute=12),
                                                                                                                     datetime(year=2016, month=11, day=11, hour=11, minute=22)]),
                                    pyAttribute(key="https://labfreed.org/terms/example/BoolAttribute", values=[True,
                                                                                                                False]),
                                    pyAttribute(key="https://labfreed.org/terms/example/ObjectAttribute", values= [{'k1':1, 'k2': {'a':'bar', 'b':'foo'}, 'k3': [0,1,2]},
                                                                                                                   {'k1':111}]),
                                    pyAttribute(key="https://labfreed.org/terms/example/Mixed", values= ["Foo",
                                                                                                         "14.88 mol/L",
                                                                                                         "HTTPS://PAC.METTORIUS.COM/-MD/CALWEIGH/A00002",
                                                                                                         datetime(year=2015, month=10, day=1, hour=10, minute=12),
                                                                                                         True,
                                                                                                         {'k1':111}])
                                ])
                            }
                )



translation_data_source = DictTranslationDataSource(
                                    supported_languages=                     ["en", "en-US", "fr"],
                                    data= Terms(terms= [
                                        Term.create("MfgDate", [    ("en", "Manufactoring date"),
                                                                    ("en-US", "Manufactoring date"),
                                                                    ("fr", "Date de fabrication")
                                                                    ]
                                        ),
                                        Term.create("MaxWeight", [    ("en", "Maximum weight"),
                                                                     ("fr", "Poids maximal")
                                                                ]
                                        ),
                                        Term.create("CalWeight", [    ("en", "Calibration weight"),
                                                                     ("fr", "Poids calibration")
                                                                ]
                                        ),
                                        Term.create("NominalWeight", [    ("en", "Nominal weight"),
                                                                     ("fr", "Poids nominel")
                                                                ]
                                        ),
                                        Term.create("https://mettorius.com/terms/attribute_group_example", [
                                            ("en", "Example Attribute Group"),
                                            ("fr", "Groupe d’attributs exemple"),
                                        ]),
                                        Term.create("https://mettorius.com/terms/attribute_group_example_lists", [
                                            ("en", "Example Attribute Group Lists"),
                                            ("fr", "Listes de groupes d’attributs exemple"),
                                        ]),
                                        Term.create("https://labfreed.org/terms/example/TextAttribute", [
                                            ("en", "Text Attribute"),
                                            ("fr", "Attribut texte"),
                                        ]),
                                        Term.create("https://labfreed.org/terms/example/NumericAttribute", [
                                            ("en", "Numeric Attribute"),
                                            ("fr", "Attribut numérique"),
                                        ]),
                                        Term.create("https://labfreed.org/terms/example/ReferenceAttribute", [
                                            ("en", "Reference Attribute"),
                                            ("fr", "Attribut de référence"),
                                        ]),
                                        Term.create("https://labfreed.org/terms/example/DateTimeAttribute", [
                                            ("en", "Date Attribute"),
                                            ("fr", "Attribut de date"),
                                        ]),
                                        Term.create("https://labfreed.org/terms/example/BoolAttribute", [
                                            ("en", "Bool Attribute"),
                                            ("fr", "Attribut booléen"),
                                        ]),
                                        Term.create("https://labfreed.org/terms/example/ObjectAttribute", [
                                            ("en", "Object Attribute (LAST RESORT)"),
                                            ("fr", "Attribut objet (EN DERNIER RECOURS)"),
                                        ]),
                                        Term.create("https://labfreed.org/terms/example/Mixed", [
                                            ("en", "Mixed List (ACTUALLY NOT ALLOWED)"),
                                            ("fr", "Liste mixte (EN FAIT INTERDITE)"),
                                        ]),

                                    ]
                                    )
                                )


server_app:Flask = AttributeServerFactory.create_server_app(datasources=[data_source, data_source_example, data_source_example_lists], 
                                                  translation_data_sources=[translation_data_source], 
                                                  authenticator= NoAuthRequiredAuthenticator(), 
                                                  default_language="en")

SERVER_URL = "127.0.0.1"
PORT = 5000

class FlaskServerThread(threading.Thread):
    def __init__(self, app, host: str, port: int):
        super().__init__(daemon=True)
        self._server = make_server(host, port, app)
        self._ctx = app.app_context()
        self._ctx.push()

    def run(self):
        self._server.serve_forever()

    def shutdown(self):
        self._server.shutdown()


client = AttributeClient(http_post_callback=http_attribute_request_default_callback_factory() )
  
def test():
    pac_id = PAC_ID.from_url("HTTPS://PAC.METTORIUS.COM/-MD/BAL500/12340")
    pac_id2 = PAC_ID.from_url("HTTPS://PAC.METTORIUS.COM/-MD/BAL500/12341*ABC$TREX/A$T.A:BLUBB")
    server = FlaskServerThread(server_app, SERVER_URL, PORT)
    server.start()
    
    try:
        attribute_groups = client.get_attributes(server_url=f"http://{SERVER_URL}:{PORT}", pac_id=pac_id)
        for ag in attribute_groups:
            print(ag.group_key)
    finally:
        server.shutdown()
        server.join(timeout=2)
    
def test2():
    pac_id = PAC_ID.from_url("HTTPS://PAC.METTORIUS.COM/-MD/BAL500/12340")
    pac_id2 = PAC_ID.from_url("HTTPS://PAC.METTORIUS.COM/-MD/BAL500/12341*ABC$TREX/A$T.A:BLUBB")

    server_app.run()
    



if __name__ == "__main__":
    test2()




from abc import ABC, abstractproperty

from pydantic import computed_field, model_validator

from labfreed.labfreed_infrastructure import LabFREED_BaseModel, ValidationMsgLevel

    
class ExtensionBase(ABC):
    name: str|None
    type: str|None


    @abstractproperty
    def data(self) -> str:
        raise NotImplementedError("Subclasses must implement 'data'")
    
    def __str__(self):
        if self.name and self.type:
            return f'{self.name}${self.type}/{self.data}'
        else:
            return self.data
    


class Extension(LabFREED_BaseModel,ExtensionBase):
    '''Implementation of Extension for unknown extension types'''
    name:str|None
    type:str|None
    data_:str
    
    @computed_field
    @property
    def data(self) -> str:
        return self.data_
       
    @staticmethod
    def create(*, name:str|None, type:str|None, data:str):
        return Extension(name=name, type=type, data=data)
    
    @model_validator(mode='before')
    @classmethod
    def move_data_field(cls, values):
        if "data" in values:
            values["data_"] = values.pop("data")
        return values
    
    model_config = {
        "extra": "allow",  # Allow extra keys during pre-validation
    }
        
    @model_validator(mode='after')
    def validate_model(self):
        if self.name and not self.type:
            raise ValueError('Extension has a name, but no type. Either set both or none')
        
        if self.type and not self.name:
            raise ValueError('Extension has a type, but no name. Either set both or none')
        
        if not self.type and not self.name:
            self._add_validation_message(msg="Extensions has no name and type. It is RECOMMENDED to specify name and type.",
                                         level=ValidationMsgLevel.RECOMMENDATION, 
                                         source=f"Extension '{self.data[0:10] if len(self.data)>10 else self.data}'",
                                         highlight_pattern=self.data)
        
        return self
            
        
    
    model_config = {
        "extra": "allow",  # Allow extra keys during pre-validation
    }
    




import re
from typing import   Any, Self
from urllib.parse import unquote
from werkzeug.datastructures import LanguageAccept
from werkzeug.http import parse_accept_header
from pydantic import ConfigDict, field_validator, model_validator
from labfreed.labfreed_infrastructure import LabFREED_BaseModel, LabFREED_ValidationError, ValidationMsgLevel
from labfreed.pac_id.pac_id import PAC_ID

ATTR_GROUPS = 'attr_grps'
ATTR_GROUPS_FWD_LKP= 'attr_fwd_lkp'


class AttributeRequestData(LabFREED_BaseModel):  
    model_config = ConfigDict(arbitrary_types_allowed=True)  
    
    pac_id: str
    language_preferences: LanguageAccept|None = None
    restrict_to_attribute_groups: list[str]|None = None
    do_forward_lookup: bool = True
    
    def as_json(self):
        return self.model_dump_json()
    
    @classmethod
    def from_json(cls, json) -> Self:
        return cls.model_validate_json(json)
    
    @classmethod
    def from_http_request(cls, pac_id:str, params:dict, headers:dict):
        # Azure seems to meddle with double slashes in a path, even if url encoded. This is to rectify this behaviour and add back a second slash if necessary
        pac_id = re.sub('HTTPS:/{1,2}', 'HTTPS://', pac_id, re.IGNORECASE)
        
        restrict_to_attribute_groups = params.get(ATTR_GROUPS)
        if restrict_to_attribute_groups == '':
            restrict_to_attribute_groups = None
        if restrict_to_attribute_groups:
            restrict_to_attribute_groups = restrict_to_attribute_groups.split(',')
            
        fwd_lkp = params.get(ATTR_GROUPS_FWD_LKP, True)
        if fwd_lkp is True:
            do_forward_lookup = True
        else:
            do_fwd_lookup =  fwd_lkp.lower() not in ['false', 'no', '0', 'n', 'off']

        lang_hdr = headers.get('Accept-Language')
        language_preferences: LanguageAccept = parse_accept_header(lang_hdr, LanguageAccept)
        out = cls(pac_id=pac_id, 
                            restrict_to_attribute_groups = restrict_to_attribute_groups,
                            do_forward_lookup = do_forward_lookup,
                            language_preferences=language_preferences
                            )   
        return out
        

    
    @model_validator(mode="before")
    @classmethod
    def _scalars_to_list(cls, d):
        if isinstance(lp:= d.get("language_preferences"), str):
            d["language_preferences"] = [lp]
        if isinstance(rag := d.get("restrict_to_attribute_groups"), str):
            d["restrict_to_attribute_groups"] = [rag]
        return d
    
    
    @field_validator('language_preferences', mode='before')
    @classmethod
    def convert_language_preferences(cls,lp):
        if isinstance(lp, LanguageAccept):
            return lp
        lq = [(lng, 1-i/len(lp)) for i, lng in enumerate(lp)]
        return LanguageAccept(lq)
        
    
    @model_validator(mode="after")
    def _revert_url_encoding(self):
        self.pac_id = unquote(self.pac_id)
        if self.restrict_to_attribute_groups:
            self.restrict_to_attribute_groups = [unquote(g) for g in self.restrict_to_attribute_groups]
        return self
           
    @model_validator(mode="after")
    def _validate_pacs(self) -> Self:           
        try:
            PAC_ID.from_url(self.pac_id)
        except LabFREED_ValidationError:
            self._add_validation_message(
                    source="pac_id",
                    level = ValidationMsgLevel.ERROR,
                    msg=f'{self.pac_id} is not a valid PAC-ID'
                )
                
        if not self.is_valid:
            raise LabFREED_ValidationError(message='Invalid request', validation_msgs=self.validation_messages())
                
        return self
    
    def language_preference_http_header(self) -> dict[str, str]:
        if not self.language_preferences:
            return {}
        headers={'Accept-Language':  LanguageAccept(self.language_preferences).to_header()}
        return headers
    
    def request_params(self) -> dict[str, Any]:
        params = {ATTR_GROUPS_FWD_LKP: self.do_forward_lookup}
        if self.restrict_to_attribute_groups:
            params.update({ATTR_GROUPS: ','.join(self.restrict_to_attribute_groups)})
        return params
        
    
    



from abc import ABC, abstractmethod, abstractproperty
import json

from pydantic import ValidationError
import rich
from labfreed.utilities.translations import Terms, Term

class TranslationDataSource(ABC):

    @abstractmethod
    def get_translations_for(self, key:str) -> Term:
        pass
    
    @abstractproperty
    def supported_languages(self) -> set[str]:
        pass
    

class DictTranslationDataSource(TranslationDataSource):
    def __init__(self, data:Terms, supported_languages:set[str]) -> None:
        
        self._data = data
        
        #check that there are translations for the supported languages. Adjust supported_languages if needed
        supported_languages = set(supported_languages) # to ensure uniqueness
        for language in supported_languages.copy():
            if translation_missing := self._list_missing_translations_to(language):
                rich.print(f"[bold yellow]WARNING:[/bold yellow]: Translation to '{language}' missing for {translation_missing}")
                supported_languages.remove(language)
                
        self._supported_languages = supported_languages
        
    def _list_missing_translations_to(self, language:str):
        translation_missing_for_keys = []
        for term in self._data.terms:
            if language not in [t.language_code for t in term.translations]:
                translation_missing_for_keys.append(term.key)
        return translation_missing_for_keys 
    
    def get_translations_for(self, key:str) -> Term:
        t = self._data.translations_for_term(key)
        return t
    
    @property
    def supported_languages(self) -> set[str]:
        return self._supported_languages
    
    
   

    
class JsonFileTranslationDataSource(DictTranslationDataSource):
    def __init__(self, path:str) -> None:
        with open(path) as f:
            data = json.load(f)
        try:
            super().__init__(data=data)
        except ValidationError as e:
            e.add_note('Json must be convertible to OnthologyTranslationDataSource')
            raise e



from abc import ABC
from datetime import  datetime, date
import re
from typing import  Annotated, Any,  Literal, Union, get_args
from urllib.parse import urlparse

from labfreed.utilities.ensure_utc_time import ensure_utc
from labfreed.labfreed_infrastructure import  LabFREED_BaseModel, ValidationMsgLevel, _quote_texts
from pydantic import   BaseModel, Field, field_validator, model_validator



class AttributeItemsElementBase(LabFREED_BaseModel, ABC):
    value: Any
    type:str
    
    @model_validator(mode="after")
    def _no_base_instances(self):
        if type(self) is AttributeItemsElementBase:
            raise TypeError("AttributeItemsElementBase must not be instantiated")
        return self
        
    # def __init__(self, **data):
    #     # Automatically inject the Literal value for `type`
    #     discriminator_value = self._get_discriminator_value()
    #     data["type"] = discriminator_value
    #     super().__init__(**data)       
    
    # @classmethod
    # def _get_discriminator_value(cls) -> str:
    #     """Extract the Literal value from the 'type' annotation."""
    #     try:
    #         type_annotation = cls.__annotations__["type"]
    #         literal_value = get_args(type_annotation)[0]
    #         return literal_value
    #     except Exception as e:
    #         raise TypeError(
    #             f"{cls.__name__} must define `type: Literal[<value>]` annotation"
    #         ) from e
            

    
class DateTimeAttributeItemsElement(AttributeItemsElementBase):
    type: Literal["datetime"] = "datetime"
    value: date | datetime
    
    @field_validator('value', mode='after')
    def set_utc__if_naive(cls, value):
        if isinstance(value, datetime):
            return ensure_utc(value)
        else:
            return value
        
    
    
class BoolAttributeItemsElement(AttributeItemsElementBase):
    type: Literal["bool"] = "bool"
    value: bool
    


    
class TextAttributeItemsElement(AttributeItemsElementBase):
    type: Literal["text"] = "text"
    value: str
    
    @model_validator(mode='after')
    def _validate_value(self):
        _validate_text(self, self.value)
        return self
       

def _validate_text(mdl:LabFREED_BaseModel, v):
    if len(v) > 5000: 
        mdl._add_validation_message(
            source="Text Attribute",
            level=ValidationMsgLevel.WARNING,  # noqa: F821
            msg=f"Text attribute {v} exceeds 5000 characters. It is recommended to stay below",
            highlight_pattern = f'{v}'
        )
            


class ReferenceAttributeItemsElement(AttributeItemsElementBase):
    type: Literal["reference"] = "reference"
    value: str 
    
          

class ResourceAttributeItemsElement(AttributeItemsElementBase):
    type: Literal["resource"] = "resource"
    value: str 
    
    @model_validator(mode='after')
    def _validate_value(self):
        _validate_resource(self, self.value)
        return self

    
def _validate_resource(mdl:LabFREED_BaseModel, v):
    r = urlparse(v)
    if not all([r.scheme, r.netloc]):
        mdl._add_validation_message(
            source="Resource Attribute",
            level=ValidationMsgLevel.ERROR,  # noqa: F821
            msg="Must be a valid url",
            highlight_pattern = f'{v}'
        )
    pattern = re.compile(r"\.\w{1,3}$", re.IGNORECASE)
    if not bool(pattern.search(v)):
        mdl._add_validation_message(
            source="Resource Attribute",
            level=ValidationMsgLevel.WARNING,  # noqa: F821
            msg="It is RECOMMENDED resource links end with a file extension",
            highlight_pattern = f'{v}'
        )
    

class NumericAttributeItemsElement(AttributeItemsElementBase):
    type: Literal["numeric"] = "numeric"
    value: str 
    _numerical_value:str
    _unit:str
    
    @model_validator(mode='after')
    def _validate_model(self):
        self._numerical_value, self._unit = self.value.split(' ', 1)
        self._validate_value()
        self._validate_unit()
        return self

    
    def _validate_value(self):
        value = self._numerical_value
        if not_allowed_chars := set(re.sub(r'[0-9\.\-\+Ee]', '', value)):
            self._add_validation_message(
                source="Numeric Attribute",
                level=ValidationMsgLevel.ERROR,  # noqa: F821
                msg=f"Characters {_quote_texts(not_allowed_chars)} are not allowed in quantity segment. Must be a number.",
                highlight_pattern = f'{value}',
                highlight_sub=not_allowed_chars
            )
        if not re.fullmatch(r'-?\d+(\.\d+)?([Ee][\+-]?\d+)?', value):
            self._add_validation_message(
                source="Numeric Attribute",
                level=ValidationMsgLevel.ERROR,
                msg=f"{value} cannot be converted to number",
                highlight_pattern = f'{value}'               
            )
    
    def _validate_unit(self):
        '''A sanity check on unit complying with UCUM. NOTE: It is not a complete validation
        - I check for blankspaces and ^, which are often used for units, but are invalid.
        - the general structure of a ucum unit is validated, but 1)parentheses are not matched 2) units are not validated 3)prefixes are not checked
        '''
        if ' ' in self._unit or '^' in self._unit:
            self._add_validation_message(
                    source="Numeric Attribute",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Unit {self._unit} is invalid. Must not contain blankspace  or '^'.",
                    highlight_pattern = self._unit
            )
        elif not re.fullmatch(r"^(((?P<unit>[\w\[\]]+?)(?P<exponent>\-?\d+)?|(?P<annotation>)\{\w+?\})(?P<operator>[\./]?)?)+", self._unit):
            self._add_validation_message(
                    source="Numeric Attribute",
                    level= ValidationMsgLevel.WARNING,
                    msg=f"Unit {self._unit} is probably invalid. Ensure it complies with UCUM specifications.",
                    highlight_pattern = self._unit
            )
    
    
    
class ObjectAttributeItemsElement(AttributeItemsElementBase):
    type: Literal["object"] = "object"
    value: dict[str, Any]
    
    
AttributeItemsElement = Annotated[
    Union[
        DateTimeAttributeItemsElement,
        BoolAttributeItemsElement,
        TextAttributeItemsElement,
        NumericAttributeItemsElement,
        ReferenceAttributeItemsElement,
        ResourceAttributeItemsElement,
        ObjectAttributeItemsElement
    ],
    Field(discriminator="type"),
]
    

           
class Attribute(LabFREED_BaseModel):
    key: str|None = Field(exclude=True)
    label: str = ""
    items: list[AttributeItemsElement]
    
     

class AttributeGroup(LabFREED_BaseModel):
    group_key: str
    group_label: str = ""
    attributes: dict[str, Attribute]
    
    @field_validator("attributes", mode="before")
    @classmethod
    def set_attribute_keys(cls, v):
        if not isinstance(v, dict):
            return v

        out = {}
        for k, a in v.items():
            if isinstance(a, dict):
                # raw input dict -> inject key if missing
                out[k] = {**a, "key": a.get("key") or k}
            else:
                # already an Attribute (or something pydantic can parse)
                out[k] = a
        return out
       


class AttributesOfPACID(LabFREED_BaseModel):
    pac_id: str
    attribute_groups: list[AttributeGroup]
    
    
    
IMPORT_URL = "https://vocab.labfreed.org/attributes/v1.jsonld"

class AttributeResponsePayload(LabFREED_BaseModel):
    schema_version: str = Field(default='1.0')
    language:str 
    data: list[AttributesOfPACID]    
    
    context: str = Field(alias='@context', default=IMPORT_URL)
      
    
    def to_json(self):
        return self.model_dump_json(exclude_none=True, by_alias=True)
    
    












from datetime import UTC, date, datetime, time
import json
import logging
from typing import  Literal
from enum import Enum
import warnings
from deprecated import deprecated
from pydantic import RootModel, field_validator, model_validator

from labfreed.labfreed_infrastructure import LabFREED_BaseModel
from labfreed.pac_attributes.api_data_models.response import (Attribute, AttributeItemsElementBase, AttributeGroup, 
                                                              BoolAttributeItemsElement, DateTimeAttributeItemsElement, NumericAttributeItemsElement,  
                                                              ObjectAttributeItemsElement, ReferenceAttributeItemsElement,  ResourceAttributeItemsElement,
                                                              TextAttributeItemsElement)
from labfreed.pac_attributes.client.client_attribute_group import ClientAttributeGroup
from labfreed.pac_id.pac_id import PAC_ID
from labfreed.trex.pythonic.quantity import Quantity



class pyReference(RootModel[str]):

    def __str__(self):
        return str(self.root)
    
class pyResource(RootModel[str]):

    def __str__(self):
        return str(self.root)


# the allowed scalar types
AllowedValue = str | bool | datetime | pyReference | pyResource | Quantity | int | float | dict | object
# homogeneous list of those
AllowedList = list[AllowedValue]

class pyAttribute(LabFREED_BaseModel):
    key:str
    label:str = ""
    values: AllowedValue | AllowedList
    
    @property
    def value_list(self):    
        '''helper function to more conveniently iterate over value elements, even if it's scalar'''   
        return self.values if isinstance(self.values, list) else [self.values]
    
    @property
    @deprecated
    def value(self):
        return self.values
    
    @model_validator(mode='before')
    def value_to_values(cls, d:dict):
        value =d.pop('value', None)
        if value is not None:
            d['values'] = value
        return d
    
    @field_validator('values', mode='before')
    def handle_one_element_list(v):
        if isinstance(v, list) and len(v)==1:
            return v[0]
        else:
            return v
        
    @field_validator('key', mode='before')
    def handle_enum_key(v):
        if isinstance(v, Enum) :
            return v.value
        else:
            return v
        
    

class pyAttributes(RootModel[list[pyAttribute]]):
    def to_payload_attributes(self) -> dict[str, Attribute]:
        out = {}
        for e in self.root:
            payload_attr = self._attribute_to_attribute_payload_type(e)
            out.update({e.key: payload_attr})
        return out
    
            
    @staticmethod        
    def _attribute_to_attribute_payload_type(attribute:pyAttribute) -> AttributeItemsElementBase:
        items = []
        for value in attribute.value_list:
            
            if isinstance(value, bool):
                items.append(BoolAttributeItemsElement(value=value))
             
            elif isinstance(value, datetime | date | time):
                if getattr(value, 'tzinfo', None) and not value.tzinfo:
                    warnings.warn(f'No timezone given for {value}. Assuming it is in UTC.')
                    value.replace(tzinfo=UTC)
                items.append(DateTimeAttributeItemsElement(value=value))
                
            elif isinstance(value, Quantity|int|float):
                if not isinstance(value, Quantity):
                    value = Quantity(value=value, unit='dimensionless')
                v = f"{value.value_as_str()} {value.unit}"
                items.append(NumericAttributeItemsElement(value=v))    
                
            elif isinstance(value, str):
                # capture quantities in the form of "100.0e5 g/L"
                if Quantity.can_convert_to_quantity(value):
                    q = Quantity.from_str_with_unit(value)
                    v = f"{q.value_as_str()} {q.unit}"
                    items.append(NumericAttributeItemsElement(value=v))      
                else:
                    items.append(TextAttributeItemsElement(value=value))

            elif isinstance(value, pyReference):
                items.append(ReferenceAttributeItemsElement(value=value.root))

            elif isinstance(value, pyResource):
                items.append(ResourceAttributeItemsElement(value=value.root))
                
            elif isinstance(value, PAC_ID):
                v = value.to_url(include_extensions=False)
                items.append(ReferenceAttributeItemsElement(value=v))
                
            else: #this covers the last resort case of arbitrary objects. Must be json serializable.
                try :
                    v = json.loads(json.dumps(value))
                    items.append(ObjectAttributeItemsElement(value=v))
                except TypeError as e:  # noqa: F841
                    raise ValueError(f'Invalid Type: {type(value)} cannot be converted to attribute. You may want to use ObjectAttribute, but would have to implement the conversion from your python type yourself.')
        
                    
        if not all(type(e) is type(items[0]) for e in items):
            logging.warning("Not all elements in items have the same type. This might cause unexpected behaviour in clients.")
        
        attr = Attribute(key=attribute.key,
                         label= attribute.label,
                         items=items)
        return attr
            
  
        
    @staticmethod
    def from_payload_attributes(attributes:dict[str, Attribute]) -> 'pyAttributes':
        out = list()
        for a in attributes.values():
            values = []
            for v in a.items:
                match v:
                    case ReferenceAttributeItemsElement() :
                        values.append(pyReference(v.value))
                        
                    case ResourceAttributeItemsElement() :
                        values.append(pyResource(v.value))
                        
                    case NumericAttributeItemsElement() :                                       
                        values.append(Quantity.from_str_value(value=v._numerical_value, unit=v._unit))

                    case BoolAttributeItemsElement() :
                        values.append(v.value)
                        
                    case TextAttributeItemsElement() :
                        values.append(v.value)
                        
                    case DateTimeAttributeItemsElement() :                    
                        values.append(v.value)
                    
                    case ObjectAttributeItemsElement() :
                        values.append(v.value)
   
            attr = pyAttribute(key=a.key, 
                            label=a.label,
                            values=values
            )
            out.append(attr )
        return out
            
            
        
class pyAttributeGroup(ClientAttributeGroup):
    attributes:dict[str,pyAttribute]
    
    @staticmethod
    def from_attribute_group(attribute_group:AttributeGroup):
        data = vars(attribute_group).copy()
        data["attributes"] = {a.key: a for a in pyAttributes.from_payload_attributes(attribute_group.attributes)}
        return pyAttributeGroup(**data)


from abc import ABC
import re
from typing import Literal
from pydantic import Field, model_validator
from labfreed.well_known_keys.unece.unece_units import unece_unit_codes
from labfreed.labfreed_infrastructure import ValidationMsgLevel
from labfreed.trex.trex_base_models import AlphanumericValue, BinaryValue, BoolValue, DateValue, ErrorValue, NumericValue, TREX_Segment, TextValue, Value



class ValueSegment(TREX_Segment, Value, ABC):
    '''@private: Abstract base class for value segments'''
    type:str
    
    @model_validator(mode='after')
    def _validate_type(self):
        valid_types = unece_unit_codes() + ['T.D', 'T.B', 'T.A', 'T.T', 'T.X', 'E']
        if self.type not in valid_types:
            self._add_validation_message(
                    source=f"TREX value segment {self.key}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Type {self.type} is invalid. Must be 'T.D', 'T.B', 'T.A', 'T.T', 'T.X', 'E' or a UNECE unit",
                    highlight_pattern = self.type
            )
        return self
        
    
    def serialize(self) -> str:
        return f'{self.key}${self.type}:{self.value}'
    
    
class NumericSegment(ValueSegment, NumericValue):
    '''Represents a TREX segment holding a numeric value'''
    type: str 
    key:str   
    value:str
    
class DateSegment(ValueSegment, DateValue):
    '''Represents a TREX segment holding a date'''
    type: Literal['T.D'] = Field('T.D', frozen=True)
    key:str   
    value:str
          
class BoolSegment(ValueSegment, BoolValue):
    '''Represents a TREX segment holding a boolean value'''
    type: Literal['T.B']  = Field('T.B', frozen=True) 
    key:str   
    value:str
class AlphanumericSegment(ValueSegment, AlphanumericValue):
    '''Represents a TREX segment holding a alphanumeric text'''
    type: Literal['T.A'] = Field('T.A', frozen=True)
    key:str   
    value:str
class TextSegment(ValueSegment, TextValue):
    '''Represents a TREX segment holding a text with arbitrary characters'''
    type: Literal['T.T'] = Field('T.T', frozen=True)
    key:str   
    value:str
    
class BinarySegment(ValueSegment, BinaryValue):
    '''Represents a TREX segment holding binary data'''
    type: Literal['T.X'] = Field('T.X', frozen=True)
    key:str   
    value:str
    
class ErrorSegment(ValueSegment, ErrorValue):
    '''Represents a TREX segment which has erroneous content'''
    type: Literal['E'] = Field('E', frozen=True)
    key:str   
    value:str
                     

def _deserialize_value_segment_from_trex_segment_str(trex_segment_str) -> ValueSegment:
    #re_scalar_pattern = re.compile(f"(?P<name>[\w\.-]*?)\$(?P<unit>[\w\.]*?):(?P<value>.*)")
    re_scalar_pattern = re.compile("(?P<name>.+?)\$(?P<unit>.+?):(?P<value>.+)")
    matches = re_scalar_pattern.match(trex_segment_str)
    if not matches:
        return None
    
    key, type_, value = matches.groups()
    
    match type_:
        case 'T.D':
            out = DateSegment(key=key, value=value, type=type_)
        case 'T.B':
            out = BoolSegment(key=key, value=value, type=type_)
        case 'T.A':
            out = AlphanumericSegment(key=key, value=value, type=type_)
        case 'T.T':
            out = TextSegment(key=key, value=value, type=type_)  
        case 'T.X':
            out = BinarySegment(key=key, value=value, type=type_) 
        case 'E':
            out = ErrorSegment(key=key, value=value, type=type_)
        case _:
            out = NumericSegment(value=value, key=key, type=type_)
                    
    return out    
    



import asyncio
from datetime import datetime, timezone
import sys
from bleak.backends.device import BLEDevice

from labfreed.pac_id.pac_id import PAC_ID
from labfreed.trex.pythonic import pyTREX
from labfreed.trex.trex import TREX
from labfreed_experimental.pac_disco.ble_uuid import PAC_Characteristics, ServiceUUID


from labfreed_experimental.pac_disco.ble_central import BlePacCentral


import logging
from pathlib import Path

TEST = False

TEST_SERVICE_UUID = "c7660001-f42b-4f69-950a-2c973260386d"

# Log file path
LOG_PATH = Path("pac_central.log")
if TEST:
    STATUS_LOG_LVL = logging.DEBUG
else:
    STATUS_LOG_LVL = logging.INFO


def setup_logging() -> tuple[logging.Logger, logging.Logger]:
    """
    Returns:
        debug_log  - very verbose logger (writes to file)
        status_log - minimal logger (writes to console only)
    """
    # root logger: no handlers, no output by default
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.handlers.clear()

    # ---- file handler: everything goes here ----
    file_handler = logging.FileHandler(LOG_PATH, encoding="utf-8")
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(
        logging.Formatter(
            "%(asctime)s [%(levelname)s] %(name)s: %(message)s",
            "%Y-%m-%d %H:%M:%S",
        )
    )
    root.addHandler(file_handler)

    # ---- console handler: only for "status" logger ----
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(
        logging.Formatter("%(message)s")  # simple console output
    )

    # verbose logger (goes to file via root)
    debug_log = logging.getLogger("pac.debug")
    debug_log.setLevel(logging.DEBUG)

    # minimal status logger (goes to file + console)
    status_log = logging.getLogger("pac.status")
    status_log.setLevel(STATUS_LOG_LVL)
    status_log.addHandler(console_handler)
    # avoid double-adding handlers if setup_logging() is called twice
    status_log.propagate = True  # still goes to file via root

    # mute third-party libraries on console (file still gets them)
    for noisy in ("bleak",):
        logging.getLogger(noisy).setLevel(logging.WARNING)

    return debug_log, status_log





SAVED_PAC_ID_FILE= Path('saved_pac_id.txt')

async def handle_pac_id(device: BLEDevice, pac_id_url: str) -> None:
    
    loop = asyncio.get_running_loop()
    
    try:
        # a nasty hack to account for BUCHI R-300 sending an experimental pac-id, which is no longer valid
        pac_id_url = pac_id_url.replace('RNR', '21') 
        pac_id = PAC_ID.from_url(pac_id_url)
        summary = pyTREX.from_trex(pac_id.get_extension("SUM").trex)
            
        sum_lines = '\n  '.join([f"{k}:  {v}" for k,v in summary.items()])  
        prompt = (
            f"\n---\nDetected PAC-ID from {device.name} ({device.address}):\n"
            f"{pac_id_url} \nSummary:\n  {sum_lines}\n\n"
            "Save this PAC-ID? [give a description if you want to save. type 'N' if you want to discard] "
        )
        answer = await loop.run_in_executor(None, input, prompt)

        if not answer.upper() == "N":
            ts = datetime.now(timezone.utc).isoformat()
            with open(SAVED_PAC_ID_FILE, "a", encoding="utf-8") as f:
                f.write(f"{pac_id_url}\t{ts}\t{answer}\n")
            print("PAC-ID saved.")
        else:
            print("PAC-ID discarded.")
            
    except Exception as e:
        pass


    


async def run_scanner() -> None:
    if TEST:
        service_uuids = None#[TEST_SERVICE_UUID]
    else:
        service_uuids = ServiceUUID.all_uuid()
        
    
    debug_log, status_log = setup_logging()

    central = BlePacCentral(
        service_uuids=service_uuids,
        pac_characteristics=PAC_Characteristics,
        on_pac_id=handle_pac_id,
        status_log=status_log,
        debug_log=debug_log
    )

    await central.run()

    # example: after run() returns (Ctrl+C) you can inspect:
    for pac_url, ts in central.get_seen_pac_ids():
        print(ts.isoformat(), pac_url)


def main() -> None:
    if sys.platform.startswith("win"):
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    asyncio.run(run_scanner())


if __name__ == "__main__":
    main()




from datetime import UTC, datetime, timedelta


from labfreed.pac_attributes.api_data_models.response import AttributeGroup
from labfreed.pac_id.pac_id import PAC_ID



class ClientAttributeGroup(AttributeGroup):
    ''' extends attribute group with info the client needs'''
    origin:str
    language:str
    value_from: datetime | None = None  
    





from functools import cached_property
from pathlib import Path
from urllib.parse import urlparse
from jinja2 import Environment, FileSystemLoader, select_autoescape
from pydantic import BaseModel, Field
from labfreed.pac_attributes.pythonic.py_attributes import pyAttribute, pyAttributeGroup, pyAttributes, pyReference, pyResource
from labfreed.pac_attributes.well_knonw_attribute_keys import MetaAttributeKeys
from labfreed.pac_cat.pac_cat import PAC_CAT
from labfreed.pac_cat.predefined_categories import PredefinedCategory
from labfreed.pac_id.pac_id import PAC_ID
from labfreed.pac_id_resolver.services import ServiceGroup, Service
from labfreed.labfreed_extended.app.formatted_print import StringIOLineBreak
from labfreed.trex.pythonic.data_table import DataTable
from labfreed.trex.pythonic.pyTREX import pyTREX
from labfreed.well_known_extensions.display_name_extension import DisplayNameExtension
from enum import Enum

class PacInfo(BaseModel):
    """A convenient collection of information about a PAC-ID"""
    pac_id:PAC_ID

    user_handovers: list[ServiceGroup] = Field(default_factory=list)
    actions: list[ServiceGroup] = Field(default_factory=list)
    attribute_groups:dict[str, pyAttributeGroup] = Field(default_factory=dict)
    
    
    # info about pac-id
    
    @cached_property
    def is_item_serialized(self) -> bool|None: #indicates if the item is at product level (e.g. BAL500), as opposed to a serialized instance thereof (e.g. BAL500 with SN 1234)
        if not isinstance(self.pac_id, PAC_CAT):
            return None
        cat = self.main_category
        if not isinstance(cat, PredefinedCategory):
            return None
        
        return cat.is_serialized

    
    @cached_property
    def pac_url(self):
        return self.pac_id.to_url(include_extensions=False)
    
    @cached_property
    def main_category(self):
        if isinstance(self.pac_id, PAC_CAT):
            return self.pac_id.categories[0]
        else:
            return None
        
        
    
    # attached data
        
    @cached_property
    def attached_data(self) -> dict[str, pyTREX]:
        return { trex_ext.name: pyTREX.from_trex(trex=trex_ext.trex) for trex_ext in self.pac_id.get_extension_of_type('TREX')}

    
    @cached_property
    def summary(self) -> pyTREX:
        return pyTREX.from_trex(self.pac_id.get_extension('SUM').trex)
    
    
    @cached_property
    def status(self) -> pyTREX:
        return pyTREX.from_trex(self.pac_id.get_extension('STATUS').trex)

        
        
        
    # Handovers and Actions
        
    def get_user_handovers_by_intent(self, intent:str, partial_match=False) -> list[Service]:
        # capture the common mistake of forgetting to access key of enum
        if isinstance(intent, Enum):
            intent = intent.value    
        services = [s for sg in self.user_handovers for s in sg.services if self._match_intent(intent, s.application_intents, partial_match)]
        return services
    
    def get_user_handover_by_intent(self, intent:str, partial_match=False, mode="first"):
        if isinstance(intent, Enum):
            intent = intent.value 
        handovers = self.get_user_handovers_by_intent(intent=intent, partial_match=partial_match)
        return self._pick_from_list(handovers, mode)
        
    
    
    def get_actions_by_intent(self, intent:str, partial_match=False) -> list[Service]:
        if isinstance(intent, Enum):
            intent = intent.value 
        actions = [s for sg in self.actions for s in sg.services if self._match_intent(intent, s.application_intents, partial_match)]
        return actions
    
    def get_action_by_intent(self, intent:str, partial_match=False, mode="first"):
        if isinstance(intent, Enum):
            intent = intent.value 
        actions = self.get_actions_by_intent(intent=intent, partial_match=partial_match)
        return self._pick_from_list(actions, mode)
    
    
    def _match_intent(self, intent, intents, partial_match):
            if partial_match:
                # intent 'document' should match 'document-operation-manual' etc
                return any([intent in i for i in intents])
            else:
                # only exact match
                return intent in intents
    
    
    @cached_property
    def important_handovers(self) -> list[Service]:
        return self.get_user_handovers_by_intent('important')
    
    
    @cached_property
    def important_actions(self) -> list[Service]:
        return self.get_actions_by_intent('important')
        
        
        
        
        
        
     # Attributes   
        
    @cached_property
    def _all_attributes(self) -> dict[str, pyAttribute]:
        out = {}
        for ag in self.attribute_groups.values():
            out.update(ag.attributes)   
        return out
    
        
    def get_attributes(self, key:str) -> list[pyAttribute]:
        # capture the common mistake of forgetting to access key of enum
        if isinstance(key, Enum):
            key = key.value    
        attributes = [a for k, a in self._all_attributes.items() if key in a.key]
        return attributes  
    
    def get_attribute(self, key:str, mode="first"):  
        # capture the common mistake of forgetting to access key of enum
        if isinstance(key, Enum):
            key = key.value      
        attributes = self.get_attributes(key)
        return self._pick_from_list(attributes, mode)

        
    def _pick_from_list(self, list, mode):
        if mode not in ['first', 'last']:
            raise ValueError('mode must be "first or "last" ')
        
        if not list:
            return None
        if mode == 'first':
            return list[0]
        if mode == 'last':
            return list[-1]
        
        
    @cached_property
    def image_url(self) -> str:
        image_attr = self._all_attributes.get(MetaAttributeKeys.IMAGE.value)
        if isinstance(image_attr.values, pyResource):
            return image_attr.values.root
        if isinstance(image_attr.values, str):
            return image_attr.values
        
        
    @cached_property
    def display_name(self) -> str|None:
        display_name = None
        pac = self.pac_id
        if dn := pac.get_extension('N'):
            dn = DisplayNameExtension.from_extension(dn)
            display_name = dn.display_name or ""
        # there can be a display name in attributes, too

        if dn_attr := self._all_attributes.get(MetaAttributeKeys.DISPLAYNAME.value): 
            dn = dn_attr.values
            display_name = dn + f' ( aka {display_name} )' if display_name else dn
            
        if not display_name and self.main_category:
            seg_240 = [s for s in self.main_category.segments if s.key=="240"]
            display_name = seg_240[0].value
            
        return display_name
    
    
    @cached_property
    def safety_pictograms(self) -> dict[str, pyAttribute]:
        pictogram_attributes = {k: a for k, a in self._all_attributes.items() if "https://labfreed.org/ghs/pictogram/" in a.key}
        return pictogram_attributes    
    
    
    @cached_property
    def qualification_state(self) -> pyAttribute:
        if state := self._all_attributes.get("https://labfreed.org/qualification/status"): 
            return state

    
         


        
    
    
    
    
    
    
    
    
########
    
    
    def format_for_print(self, markup:str='rich') -> str:
        
        printout = StringIOLineBreak(markup=markup)
        
        printout.write(f"for {self.pac_url}")
        
        printout.title1("Info")
        printout.key_value("Display Name", self.display_name)
        
        if isinstance(self.pac_id, PAC_CAT):
            printout.title1("Categories")
            for c in self.pac_id.categories:
                category_name = c.__class__.__name__
                printout.title2(category_name)
                for k,v in c.segments_as_dict().items():
                    printout.key_value(k, v)
                
                    
        printout.title1("Services")
        for sg in self.user_handovers:           
            printout.title2(f"(from {sg.origin})")
            for s in sg.services:
                printout.link(s.service_name, s.url)          
        
        
        printout.title1("Attributes")
        for ag in self.attribute_groups.values():  
            printout.title2(f'{ag.group_label} (from {ag.origin})')
            for v in ag.attributes.values():
                v:pyAttribute
                #print(f'{k}: ({v.label})           :: {v.value}  ')
                printout.key_value(v.label, ', '.join([str(e) for e in v.value_list]))
      
        out =  printout.getvalue()

        return out
    
    
    

    
    


from enum import Enum


class ServiceUUID(Enum):
    MATERIAL_DEVICE = "6b190a7d-5aac-4943-bc3f-0ca90a9dff66"
    MATERIAL_SUBSTANCE = "5a0f2312-e944-4761-adf1-ecac1ea30171"
    MATERIAL_CONSUMABLE = "72e77869-da3c-4dab-b540-563ba5085a3e"
    MATERIAL_MISC = "971b275e-4fa1-4993-bdc6-720d159ffb32"
    
    DATA_RESULT = "1b79bd00-b4f9-4932-b93b-17b9ee161d0c"
    DATA_METHOD = "640cf34a-6a93-423b-adc4-9836d24f33d1"
    DATA_CALIBRATION = "98c5ef48-2059-4692-ac8d-607c97afeb1d"
    DATA_PROGRESS = "114c07af-9d07-48bc-adc0-de34d1d075ce"
    DATA_MISC = "2002cf5a-569f-4d37-84ff-19aeeda6fc72"
    
    PROCESSOR_SOFTWARE = "76a5901e-2438-4f1d-9738-e7c35c813c8c"
    PROCESSOR_MISC = "0e067884-53a4-4470-a5f8-02dec4d4e171"
    
    @classmethod
    def all_uuid(cls):
        out = [member.value.lower() for member in cls]
        return out
        
    
    
class PAC_Characteristics(Enum):
    PAC_ID_PART_0 = "1b79bd10-b4f9-4932-b93b-17b9ee161d0c"
    PAC_ID_PART_1 = "f636bdc7-88da-49e5-b823-12310e62c215"
    PAC_ID_PART_2 = "644c3de5-4337-44be-9dec-8b3abf10bf70"
    PAC_ID_PART_3 = "382bfd6a-a407-48d3-9337-1b1b81a2c314"


## Materials
from abc import ABC, abstractproperty
from pydantic import Field, computed_field, model_validator

from labfreed.labfreed_infrastructure import ValidationMsgLevel
from labfreed.pac_cat.category_base import Category
from labfreed.pac_id.id_segment import IDSegment

class PredefinedCategory(Category, ABC):
    '''@private 
    Base for Predefined catergories
    '''
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    
    @computed_field
    @property
    def segments(self) -> list[IDSegment]:
        return self._get_segments(use_short_notation=False)
    
    @abstractproperty
    def is_serialized(self) -> bool:
        pass
    
    def _get_segments(self, use_short_notation=False) -> list[IDSegment]:
        segments = []
        can_omit_keys = use_short_notation # keeps track of whether keys can still be omitted. That is the case when the segment recommendation is followed
        for field_name, field_info in self.model_fields.items():
            if field_name in ['key', 'additional_segments']:
                continue
            if value := getattr(self, field_name):
                if can_omit_keys:
                    key = None
                else:
                    key = field_info.alias
                segments.append(IDSegment(key= key, value= value)  )
            else:
                can_omit_keys = False
        if self.additional_segments:
            segments.extend(self.additional_segments)
        return segments
    
    model_config = {
        "populate_by_name": True 
    }
    ''' @private Pydantic tweak to allows model fields to be populated using their Python name, even if they have an alias defined. 
        The alias we need to use the GS1 code in serialization
    '''



class Material_Device(PredefinedCategory):
    '''Represents the -MD category'''
    key: str =         Field(default='-MD', frozen=True)
    model_number: str|None =         Field(              alias='240')
    serial_number: str|None =        Field(              alias='21')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    
    @model_validator(mode='after')
    def _validate_mandatory_fields(self):
        if not self.model_number:
            self._add_validation_message(
                    source=f"Category {self.key}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f'Category key {self.key} is missing mandatory field Model Number',
                    highlight_pattern = f"{self.key}"
            )
        if not self.serial_number:
            self._add_validation_message(
                    source=f"Category {self.key}",
                    level = ValidationMsgLevel.WARNING,
                    msg=f'Category key {self.key} is missing field Serial Number. Check that you are indeed to a product and not a specific device.',
                    highlight_pattern = f"{self.key}"
            )
        return self
    
    @property
    def is_serialized(self) -> bool:
        return bool(self.serial_number)
    
    
class Material_Substance(PredefinedCategory):
    '''Represents the -MS category'''
    key: str =         Field(default='-MS', frozen=True)
    product_number:str|None =    Field(              alias='240')
    batch_number:str|None =     Field(default=None, alias='10')
    container_size:str|None =   Field(default=None, alias='20')
    container_number:str|None = Field(default=None, alias='21')
    aliquot:str|None =          Field(default=None, alias='250')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    
    @model_validator(mode='after')
    def _validate_mandatory_fields(self):
        if not self.product_number:
            self._add_validation_message(
                    source=f"Category {self.key}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f'Category key {self.key} is missing mandatory field Product Number',
                    highlight_pattern = f"{self.key}"
            )
        return self
    
    @property
    def is_serialized(self) -> bool:
        return bool(self.batch_number or self.container_number or self.aliquot)
    
    
class Material_Consumable(PredefinedCategory):
    '''Represents the -MC category'''
    key: str = Field(default='-MC', frozen=True)
    product_number:str|None =   Field(              alias='240')
    batch_number:str|None =     Field(default=None, alias='10')
    packaging_size:str|None =     Field(default=None, alias='20')
    serial_number:str|None =    Field(default=None, alias='21')
    aliquot:str|None =          Field(default=None, alias='250')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    
    @model_validator(mode='after')
    def _validate_mandatory_fields(self):
        if not self.product_number:
            self._add_validation_message(
                    source=f"Category {self.key}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f"Category key {self.key} is missing mandatory field 'Product Number'",
                    highlight_pattern = f"{self.key}"
            )
        return self
    
    @property
    def is_serialized(self) -> bool:
        return bool(self.batch_number or self.serial_number or self.aliquot)
    
    
class Material_Misc(Material_Consumable):
    '''Represents the -MX category'''
    # same fields as Consumable
    key: str = Field(default='-MX', frozen=True)
    product_number:str|None =   Field(              alias='240')
    batch_number:str|None =     Field(default=None, alias='10')
    packaging_size:str|None =     Field(default=None, alias='20')
    serial_number:str|None =    Field(default=None, alias='21')
    aliquot:str|None =          Field(default=None, alias='250')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    

    


## Data
class Data_Abstract(PredefinedCategory, ABC):
    '''@private'''
    key: str
    id:str|None =                    Field(              alias='21')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    
    @model_validator(mode='after')
    def _validate_mandatory_fields(self):
        if not self.id:
            self._add_validation_message(
                    source=f"Category {self.key}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f"Category key {self.key} is missing mandatory field 'ID'",
                    highlight_pattern = f"{self.key}"
            )
        return self
    
    @property
    def is_serialized(self) -> bool:
        return True
    

class Data_Result(Data_Abstract):
    '''Represents the -DR category'''
    key: str = Field(default='-DR', frozen=True)
    id:str|None =                    Field(              alias='21')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    
    
class Data_Method(Data_Abstract):
    '''Represents the -DM category'''
    key: str = Field(default='-DM', frozen=True)
    id:str|None =                    Field(              alias='21')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    
    
class Data_Calibration(Data_Abstract):
    '''Represents the -DC category'''
    key: str = Field(default='-DC', frozen=True)
    id:str|None =                    Field(              alias='21')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    
    
class Data_Progress(Data_Abstract):
    '''Represents the -DP category'''
    key: str = Field(default='-DP', frozen=True)
    id:str|None =                    Field(              alias='21')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    
    
class Data_Static(Data_Abstract):
    '''Represents the -DS category'''
    key: str = Field(default='-DS', frozen=True)
    id:str|None =                    Field(              alias='21')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    
class Data_Misc(Data_Abstract):
    '''Represents the -DX category'''
    key: str = Field(default='-DX', frozen=True)
    id:str|None =                    Field(              alias='21')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    



class Processor_Abstract(PredefinedCategory, ABC):
    '''@private'''
    key: str
    processor_instance:str|None =                    Field(              alias='21')
    processor_code:str|None =                        Field(              alias='240')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    
    @model_validator(mode='after')
    def _validate_mandatory_fields(self):
        if not self.processor_instance:
            self._add_validation_message(
                    source=f"Category {self.key}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f"Category key {self.key} is missing mandatory field 'processor instance'",
                    highlight_pattern = f"{self.key}"
            )
        return self
    
    @property
    def is_serialized(self) -> bool:
        return bool(self.processor_instance)
    

class Processor_Software(Processor_Abstract):
    '''Represents the -PS category'''
    key: str = Field(default='-PS', frozen=True)
    processor_instance:str|None =                    Field(              alias='21')
    processor_code:str|None =                        Field(              alias='240')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''

class Processor_Misc(Processor_Abstract):
    '''Represents the -PX category'''
    key: str = Field(default='-PX', frozen=True)
    processor_instance:str|None =                    Field(              alias='21')
    processor_code:str|None =                        Field(              alias='240')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    
    
    
class Misc(Category, ABC):
    '''@private'''
    key: str = Field(default='-X', frozen=True)
    id:str|None =                    Field(              alias='21')
    additional_segments: list[IDSegment] = Field(default_factory=list, exclude=True)
    ''' Category segments, which are not defined in the specification'''
    
    @model_validator(mode='after')
    def _validate_mandatory_fields(self):
        if not self.id:
            self._add_validation_message(
                    source=f"Category {self.key}",
                    level = ValidationMsgLevel.ERROR,
                    msg=f"Category key {self.key} is missing mandatory field 'ID'",
                    highlight_pattern = f"{self.key}"
            )
        return self
    
    @property
    def is_serialized(self) -> bool:
        return bool(self.id)
    
    
    
category_key_to_class_map  = {
        '-MD': Material_Device,
        '-MS': Material_Substance,
        '-MC': Material_Consumable,
        '-MX': Material_Misc,
        '-DM': Data_Method,
        '-DR': Data_Result,
        '-DC': Data_Calibration,
        '-DP': Data_Progress,
        '-DS': Data_Static,
        '-DX': Data_Misc,
        '-PS': Processor_Software,
        '-PX': Processor_Misc,
        '-X':Misc
}


from enum import Enum


class WellKnownKeys(Enum):
    GTIN = '01'
    BATCH = '10'
    VARIANT = '20'
    SERIAL = '21'
    ADDITIONAL_IDENTIFIER = '240'
    SECONDARY_SERIAL = '250'
    RUN_ID_ABSOLUTE = 'RNR'
    SAMPLE_ID = 'SMP'
    EXPERIMENT_ID = 'EXP'
    RESULT_ID = 'RST'
    METHOD_ID = 'MTD'
    REPORT_ID = 'RPT'
    TIMESTAMP = 'TS'
    VERSION = 'V'
    
    MAXWEIGHT = "MAXWEIGHT"
    LASTCALIBRATION = "LASTCAL"
    NOMINALWEIGHT = "NOMINALWEIGHT"
    
    
    def as_url(self) -> str:
        # prefix the enum’s name to point to the description in the web
        return f"labfreed.org/wkk/{self.name}"
    



from functools import cache
import json
from pathlib import Path




@cache
def unece_units() -> list[dict]:
    p = Path(__file__).parent / 'UneceUnits.json'
    with open(p) as f:
        l = json.load(f)  # noqa: E741
    return l

@cache
def unece_unit_codes():
    codes= [u.get('commonCode') for u in unece_units() if u.get('state') == 'ACTIVE']
    return codes


def unece_unit(unit_code):
    unit =  [u for u in unece_units() if u['commonCode'] == unit_code]
    if len(unit) == 0:
        return None
    else:
        return unit[0]
    
def unit_symbol(unit:dict) ->str:
    symbol:str = unit.get('symbol')
    return symbol

def unit_name(unit:dict) ->str:
    return unit.get('name')



# def check_compatibility_unece_quantities():
#     unece = unece_units()
#     print(f'Number of units in file: {len(unece)}')
    
#     failed = list()
#     success = list()
#     for u in unece:
#         if u.get('state') ==  'ACTIVE':
#             try:
#                 if not u.get('symbol'):
#                     assert False
#                 u.get('name')  
#                 validate_unit(u.get('symbol'))
#                 success.append(u)
#             except AssertionError:
#                 failed.append(u)
#         else:
#             pass
        
    
    
#     print('[blue] FAILED [/blue]')
#     for u in failed:
#         print(f'{u.get('commonCode')}: {u.get('name')}')
    
#     print('[yellow] SUCCESSFUL [/yellow]')
#     for u in success:
#         print(u)
        
#     print(f'{len(failed)} / {len(unece)} failed to convert')
        

        


from urllib.parse import parse_qs, urlencode, urlparse, urlunparse


def add_ga_params(url:str, source:str ) -> str:
    parts = urlparse(url)
    query = parse_qs(parts.query)
    query.update({
        'utm_source': f'PAC.{source}',
        'utm_medium': 'link',
        'utm_campaign': 'handover'
    })
    return urlunparse(parts._replace(query=urlencode(query, doseq=True)))


def add_trace_id_params(url:str, trace_id:str ) -> str:
    parts = urlparse(url)
    query = parse_qs(parts.query)
    query.update({
        'trace_id': trace_id
    })
    return urlunparse(parts._replace(query=urlencode(query, doseq=True)))




from enum import Enum


class MetaAttributeKeys(Enum):
    DISPLAYNAME = "https://schema.org/name"
    IMAGE = "https://schema.org/image"
    ALIAS = "https://schema.org/alternateName"
    DESCRIPTION = "https://schema.org/description"
    GROUPKEY = "https://labfreed.org/terms/attribute_group_metadata"
    
    NAME = "https://schema.org/name"
    PHONE = 'https://schema.org/telephone'
    EMAIL = 'https://schema.org/email'
    ADDRESS = 'https://schema.org/address'
    COUNTRY = 'https://schema.org/addressCountry'
    
    


from labfreed.pac_attributes.server.server import AttributeServerRequestHandler  # noqa: F401
from labfreed.pac_attributes.server.attribute_data_sources import Dict_DataSource  # noqa: F401
from labfreed.pac_attributes.server.translation_data_sources import DictTranslationDataSource  # noqa: F401



from functools import lru_cache
import logging
from typing import Self
from requests import get
from deprecated import deprecated

from labfreed.pac_cat.pac_cat import PAC_CAT
from labfreed.pac_id.pac_id import PAC_ID
from labfreed.pac_id_resolver.services import ServiceGroup
from labfreed.pac_id_resolver.cit_v1 import CIT_v1
from labfreed.pac_id_resolver.resolver_config import ResolverConfig



''' Configure pdoc'''
__all__ = ["PAC_ID_Resolver"]

def load_cit(path):
    with open(path, 'r') as f:
        s = f.read()
        return cit_from_str(s)

    
@deprecated("cit version 1 is deprecated. use resolvber config and load with ResolverConfig.from_yaml(s)")
def cit_from_str(s:str, origin:str='') -> CIT_v1|ResolverConfig:
    try:
        cit2 = ResolverConfig.from_yaml(s)
        cit_version = 'v2'
    except Exception as e1:
        cit2 = None
    try:
        cit1 = CIT_v1.from_csv(s, origin)
        cit_version = 'v1'  # noqa: F841
    except Exception as e2:
        cit1 = None
    
    cit = cit2 or cit1 or None
    return cit

@lru_cache
def _get_issuer_resolver_config(issuer:str):
    '''Gets the issuer's cit.'''
    # V2
    url = 'HTTPS://PAC.' + issuer + '/resolver_config.yaml'
    try:
        r = get(url, timeout=2)
        if r.status_code < 400:
            config_str = r.text
            resolver_config = ResolverConfig.from_yaml(config_str)
            return resolver_config
        else: 
            logging.error(f"Could not get CIT V2 form {issuer}")
    except Exception:
        logging.error(f"Could not get CIT V2 form {issuer}")
        
        
    # V1 (as fallback)
    url = 'HTTPS://PAC.' + issuer + '/coupling-information-table'
    try:
        r = get(url, timeout=2)
        if r.status_code < 400:
            config_str = r.text
            cit = CIT_v1.from_csv(config_str, '')
            return cit
        else: 
            logging.error(f"Could not get CIT form {issuer}")
    except Exception:
        logging.error(f"Could not get CIT form {issuer}")
        
    # no cit found
    return None 

    


class PAC_ID_Resolver():
    def __init__(self, resolver_configs:list[ResolverConfig|CIT_v1]=None) -> Self:
        '''Initialize the resolver with coupling information tables'''
        if not resolver_configs:
            resolver_configs = []
        self._resolver_configs = set(resolver_configs)
            
        
    def resolve(self, pac_id:PAC_ID|str, check_service_status=True, use_issuer_resolver_config=True) -> list[ServiceGroup]:
        '''Resolve a PAC-ID'''
        if isinstance(pac_id, str):
            pac_id_catless = PAC_ID.from_url(pac_id, try_pac_cat=False)
            pac_id = PAC_CAT.from_url(pac_id)
        
        # it's likely to
        if isinstance(pac_id, PAC_ID):
            pac_id_catless = PAC_ID.from_url(pac_id.to_url(), try_pac_cat=False)
        else:
            raise ValueError('pac_id is invalid. Should be a PAC-ID in url form or a PAC-ID object')
    
                
        resolver_configs = self._resolver_configs.copy()
        if use_issuer_resolver_config:
            if issuer_resolver_config := _get_issuer_resolver_config(pac_id.issuer):
                resolver_configs.add(issuer_resolver_config)
         
        matches = []
        for cit in resolver_configs:
            try:
                if isinstance(cit, CIT_v1):
                    # cit v1 has no concept of categories and implied keys. It would treat these segments as value segment
                    matches.append(cit.evaluate_pac_id(pac_id_catless))
                else:
                    matches.append(cit.evaluate_pac_id(pac_id))
            except Exception as e:
                logging.error(f'Failed to resolve pac {pac_id.to_url()} with cit {cit.origin}')
        
        if check_service_status:
            for m in matches:
                m.update_states()   
        return matches
            
    
    
if __name__ == '__main__':
    r = PAC_ID_Resolver()
    r.resolve()



import asyncio
import logging
import sys
from datetime import datetime, timezone
from typing import Awaitable, Callable, Iterable, Dict

from bleak import BleakClient, BleakScanner
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

from labfreed.pac_id.pac_id import PAC_ID
from labfreed_experimental.pac_disco.ble_uuid import PAC_Characteristics, ServiceUUID


TEST = True
TEST_SERVICE_UUID = "c7660001-f42b-4f69-950a-2c973260386d"

RSSI_THRESHOLD = -80

PacIdCallback = Callable[[BLEDevice, str], Awaitable[None]]


class BlePacCentral:
    """
    BLE central that scans for devices, connects, reads PAC_ID characteristics,
    tracks seen PAC IDs (with timestamps), and calls an injected async callback
    for each *new* PAC ID.
    """

    def __init__(
        self,
        *,
        service_uuids: Iterable[str],
        pac_characteristics: Iterable[PAC_Characteristics],
        on_pac_id: PacIdCallback,
        debug_log: logging.Logger, status_log: logging.Logger
    ) -> None:
        self._service_uuids = list(service_uuids)
        self._pac_characteristics = list(pac_characteristics)
        self._on_pac_id = on_pac_id
        self._log = debug_log
        self._status = status_log

        # pac_id_url -> first_seen_timestamp (UTC)
        self._seen_pac_ids: Dict[str, datetime] = {}

        # per-device lock to avoid overlapping connects/reads for same address
        self._per_device_lock: Dict[str, asyncio.Lock] = {}

        self._scanner: BleakScanner | None = None

    # ------- public API -------

    async def run(self) -> None:
        """Run scanner until Ctrl+C."""
        self._scanner = BleakScanner(
            detection_callback=self._detection_callback,
            service_uuids=self._service_uuids,
            scanning_mode="active",
        )

        await self._scanner.start()
        self._log.info("Scanning for BLE advertisements... Ctrl+C to stop.")
        try:
            while True:
                await asyncio.sleep(1.0)
        except KeyboardInterrupt:
            self._log.info("Stopping scanner...")
        finally:
            await self._scanner.stop()

    def get_seen_pac_ids(self) -> list[tuple[str, datetime]]:
        """
        Synchronous snapshot of all seen PAC IDs.

        Returns list of (pac_id_url, first_seen_utc) tuples.
        """
        return list(self._seen_pac_ids.items())

    # ------- internal helpers -------

    def _get_lock_for(self, address: str) -> asyncio.Lock:
        lock = self._per_device_lock.get(address)
        if lock is None:
            lock = asyncio.Lock()
            self._per_device_lock[address] = lock
        return lock

    def _detection_callback(self, device: BLEDevice, adv: AdvertisementData) -> None:
        msg = f"Detected ({device.address}) aka ({device.name}). Services:{adv.service_uuids}, RSSI:{adv.rssi} "
        self._log.info(msg)
        
        # ignore weak signals (far away)
        if adv.rssi is not None and adv.rssi < RSSI_THRESHOLD:
            return

        # only try connectable devices if info is present
        if hasattr(adv, "is_connectable") and not adv.is_connectable:
            return

        # schedule handling; per-device lock will serialize connects per address
        asyncio.create_task(self._handle_device(device))

    async def _handle_device(self, device: BLEDevice) -> None:
        lock = self._get_lock_for(device.address)
        async with lock:
            pac_id_url = await self._read_pac_id_url(device)
            if not pac_id_url:
                return

            # de-duplicate by PAC ID (not by device)
            if pac_id_url in self._seen_pac_ids:
                self._log.info(f"PAC-ID {pac_id_url} was already known")
                return

            # record first seen timestamp
            self._seen_pac_ids[pac_id_url] = datetime.now(timezone.utc)

            # fire user callback
            await self._on_pac_id(device, pac_id_url)

    async def _read_pac_id_url(self, device: BLEDevice) -> str:
        self._log.info(f"Connecting to {device.address} ({device.name})...")
        pac_id_url = ""

        try:
            # services_cache=False helps on Windows to avoid stale service info
            async with BleakClient(device) as client:
                for characteristic in self._pac_characteristics:
                    try:
                        raw = await client.read_gatt_char(characteristic.value)
                        if raw:
                            pac_id_url += raw.decode("utf-8")
                    except Exception as e:
                        self._log.warning(
                            f"{device.address} cannot read {characteristic.name}: {e!r}"
                        )

            if pac_id_url:
                self._log.info(f"PAC_ID {pac_id_url} was published")

        except Exception as e:
            self._log.error(f"Failed to connect to {device.address}: {e!r}")

        return pac_id_url



from enum import Enum, auto
import logging
import re
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr, field_validator
from typing import Any, List, Set

import warnings
import functools
import inspect

from rich import print
from rich.table import Table

''' Configure pdoc'''
__all__ = ["LabFREED_BaseModel", "ValidationMessage", "ValidationMsgLevel", "LabFREED_ValidationError"]

class PDOC_Workaround_Base(BaseModel):
    '''@private
    This class only exists to make pdoc work better with Pydantic models. It is set up such, that some things are not showing up in docu
    '''
    model_config = ConfigDict(extra="forbid")
    """@private"""
    def model_post_init(self, context: Any) -> None:
        '''@private'''
        super().model_post_init(context)





class ValidationMsgLevel(Enum):
    '''
    Level of validation messages
    '''
    ERROR = auto() 
    '''Model is **invalid**'''
    WARNING = auto() 
    '''Model is **valid**, but has issues'''
    RECOMMENDATION = auto()
    '''Model is **valid**, but recommendations apply'''
    INFO = auto()
    '''Model is **valid**. Something of interest was detected, which is not a recommendation.'''

class ValidationMessage(PDOC_Workaround_Base):
    '''
    Represents one problem in the model
    '''
    source_id:int
    source:str 
    level: ValidationMsgLevel
    msg:str
    highlight:str = "" #this can be used to highlight problematic parts
    highlight_sub_patterns:list[str] = Field(default_factory=list)

    @field_validator('highlight_sub_patterns', mode='before')
    @classmethod
    def _ensure_list(cls, v):
        if isinstance(v, str):
            return [v]
        return v


    
class LabFREED_ValidationError(ValueError):
    '''Error which is raised, when LabFREED validation fails, i.e. when the model contains at least one error.'''
    
    def __init__(self, message=None, validation_msgs=None):
        '''@private'''
        super().__init__(message)
        self._validation_msgs = validation_msgs

    @property
    def validation_msgs(self):
        ''' The validation messages (errors, recommendations, info) present in the invalid model'''
        return self._validation_msgs
    
    


class LabFREED_BaseModel(PDOC_Workaround_Base):
    """ Extension of Pydantic BaseModel, so that validator can issue warnings.
    The purpose of that is to allow only minimal validation but on top check for stricter recommendations"""
    
    _validation_messages: list[ValidationMessage] = PrivateAttr(default_factory=list)
    """Validation messages for this model"""
    
    @property
    def is_valid(self) -> bool:
        return len(self.errors()) == 0
    

    def validation_messages(self, nested=True) -> list[ValidationMessage]:
        if nested:
            msgs = self._get_nested_validation_messages()
        else:
            msgs = self._validation_messages
        return msgs
    
    def errors(self, nested=True) -> list[ValidationMessage]: 
        return _filter_errors(self.validation_messages(nested=nested))
    
    def warnings(self, nested=True) -> list[ValidationMessage]: 
        return _filter_warnings(self.validation_messages(nested=nested))
    

    def _add_validation_message(self, *, msg: str, level:ValidationMsgLevel, source:str="", highlight_pattern="", highlight_sub=None):
        if not highlight_sub:
            highlight_sub = []
        w = ValidationMessage(msg=msg, source=source, level=level, highlight=highlight_pattern, highlight_sub_patterns=highlight_sub, source_id=id(self))

        if w not in self._validation_messages:
            self._validation_messages.append(w)

    # Function to extract warnings from a model and its nested models
    def _get_nested_validation_messages(self, parent_name: str = "", visited: Set[int]|None = None) -> List['ValidationMessage']:
        """
        Recursively extract warnings from a Pydantic model and its nested fields, including computed fields.

        :param parent_name: The name of the parent model to track the path.
        :param visited: Set of visited object IDs to prevent infinite loops.
        :return: List of ValidationMessages from this and nested models.
        """
        if visited is None:
            visited = set()

        model_id = id(self)
        if model_id in visited:
            return []
        visited.add(model_id)

        warnings_list = [warning for warning in self.validation_messages(nested=False)]

        # Traverse regular fields
        for field_name, field in self.__fields__.items():
            full_path = f"{parent_name}.{field_name}" if parent_name else field_name
            value = getattr(self, field_name)

            if isinstance(value, LabFREED_BaseModel):
                warnings_list.extend(value._get_nested_validation_messages(full_path, visited))
            elif isinstance(value, list):
                for index, item in enumerate(value):
                    if isinstance(item, LabFREED_BaseModel):
                        list_path = f"{full_path}[{index}]"
                        warnings_list.extend(item._get_nested_validation_messages(list_path, visited))

        # Traverse computed fields
        mdl:BaseModel = getattr(self, '__pydantic_decorators__', {})
        computed_fields = mdl.computed_fields or {}
        for field_name in computed_fields:
            full_path = f"{parent_name}.{field_name}" if parent_name else field_name
            try:
                value = getattr(self, field_name)
            except Exception:
                continue  # Safely skip computed properties that raise errors

            if isinstance(value, LabFREED_BaseModel):
                warnings_list.extend(value._get_nested_validation_messages(full_path, visited))
            elif isinstance(value, list):
                for index, item in enumerate(value):
                    if isinstance(item, LabFREED_BaseModel):
                        list_path = f"{full_path}[{index}]"
                        warnings_list.extend(item._get_nested_validation_messages(list_path, visited))

        return warnings_list
        

    
    def print_validation_messages(self, target='console'):
        msgs = self.format_validation_messages(target=target)
        
        table = Table(title="Validation Results", show_header=False, title_justify='left')

        def col(s):
            return table.add_column(s, vertical='top')
        col("-")
        
        if not msgs:
            table.add_row('All clear!', end_section=True)
            return

        for m in msgs:
            table.add_row(m)
            table.add_section()

        logging.info(table)
        print(table)
        return table
        
        
    def format_validation_messages(self, target='console') -> list[str]:
        """Format validation messages

        Args:
            target (str, optional): Target format: 'markdown', 'console', 'html', 'html_styled'.

        Returns:
            list[str]: formated messages
        """
        formatted_msg = list()
        for m in self.validation_messages():
            if m.level == ValidationMsgLevel.ERROR:
                color = '#d70000'
            else:
                color = '#d78700'
                
            match target:
                case 'markdown':
                    fmt = lambda s: f'👉{s}👈'  # noqa: E731
                    fmt_title = lambda s: f'**{s}**'  # noqa: E731
                    br = '\n'
                case 'console':     
                    fmt = lambda s: f'[{color} bold]{s}[/{color} bold]'  # noqa: E731
                    fmt_title = fmt
                    br = '\n'
                case 'html':
                    lvl = m.level.name.lower()
                    fmt = lambda s: f'<span class="validation-problem {lvl}">{s}</span>'  # noqa: E731
                    fmt_title = lambda s: f'<span class="validation-title {lvl}">{s}</span>'  # noqa: E731
                    br = '<br>'
                case 'html_styled':
                    fmt = lambda s: f'<b style="color:{color}">{s}</b>'  # noqa: E731
                    fmt_title = fmt
                    br = '<br>'
                
            serialized = str(self)
            serialized = serialized.replace('\n', br)
            emphazised_highlight = self._emphasize_in(m, serialized, fmt=fmt)
            emphazised_highlight = emphazised_highlight.replace('👈👉','') # removes two consecutive markers, to make it cleaner
            
            txt =   f'{fmt_title(m.level.name)} in {m.source}'
            txt += br + f'{m.msg}'
            txt += br+br + emphazised_highlight
            
            formatted_msg.append(txt)
        return formatted_msg
    
    def _emphasize_in(self, validation_msg, validation_node_str:str, fmt):
        if validation_msg.highlight_sub_patterns:
            replacements = validation_msg.highlight_sub_patterns
        else:
            replacements = [validation_msg.highlight]
        # Sort patterns by length descending to avoid subpattern clobbering
        sorted_patterns = sorted(replacements, key=len, reverse=True)
        # Escape the patterns for regex safety
        escaped_patterns = [re.escape(p) for p in sorted_patterns]
        # Create one regex pattern with alternation (longest first)
        pattern = re.compile("|".join(escaped_patterns))
        
        out = pattern.sub(lambda m: fmt(m.group(0)), validation_node_str)
        return out
            

    
def _filter_errors(val_msg:list[ValidationMessage]) -> list[ValidationMessage]:
    return [ m for m in val_msg if m.level == ValidationMsgLevel.ERROR ]

def _filter_warnings(val_msg:list[ValidationMessage]) -> list[ValidationMessage]:
    return [ m for m in val_msg if m.level != ValidationMsgLevel.ERROR  ]     

def _quote_texts(texts:list[str]):
    return ','.join([f"'{t}'" for t in texts])












import re
import string
import traceback
import warnings

import rich
from werkzeug.datastructures.accept import LanguageAccept

from labfreed.pac_attributes.api_data_models.request import AttributeRequestData
from labfreed.pac_attributes.api_data_models.response import AttributeResponsePayload, AttributesOfPACID, ReferenceAttributeItemsElement
from labfreed.pac_attributes.api_data_models.server_capabilities_response import ServerCapabilities
from labfreed.pac_attributes.server.attribute_data_sources import AttributeGroupDataSource
from labfreed.pac_attributes.server.translation_data_sources import TranslationDataSource
from labfreed.pac_id.pac_id import PAC_ID


class InvalidRequestError(ValueError):
    pass


    
class AttributeServerRequestHandler():
    def __init__(self, data_sources:list[AttributeGroupDataSource], translation_data_sources:list[TranslationDataSource], default_language:str):
        '''Initializes the AttributeServerRequestHandler.
        - does some validation on availability of translations. NOTE: if data_sources or translation_data_sources change, this validation might get outdated'''
        if isinstance(data_sources, AttributeGroupDataSource):
            data_sources = [data_sources]
        self._attribute_group_data_sources: list[AttributeGroupDataSource] = data_sources
        
        
        
        # find which languages the sources consistently know
        self._translation_data_sources: list[TranslationDataSource] = translation_data_sources
        supported_languages = set()
        for tds in self._translation_data_sources:
            supported_languages.update(tds.supported_languages)
        if not supported_languages:
            raise ValueError("translation_data_sources contain no common fully supported language")
        
        if default_language not in supported_languages:
            raise ValueError(f"fallback language {default_language} is not supported by all translation data sources.")
        self._supported_languages = supported_languages
        self._default_language = default_language
        
        # check there are translations for all provided attributes
        missing_translations = []
        provided_attributes = [attr for ds in self._attribute_group_data_sources for attr in ds.provides_attributes]
        for language in self._supported_languages:
            for attribute_key in provided_attributes:
                if not self._get_display_name_for_key(attribute_key, language):
                    missing_translations.append((attribute_key, language))
        if missing_translations:
            rich.print('[yellow bold]WARNING: Missing translations[/yellow bold]')
            for mt in missing_translations:
                rich.print(f"[yellow]WARNING:[/yellow] '{mt[1]}' translation missing for '{mt[0]}'' ")
                    
               
        
    def handle_attribute_request(self, request_data:AttributeRequestData) -> str:
        if not isinstance(request_data, AttributeRequestData):
            raise ValueError('request_data most be of AttributeRequestData')
        try:
            r = AttributeRequestData.model_validate(request_data)
        except Exception as e:
            print(e)
            raise InvalidRequestError
        attributes_for_pac_id = []
        referenced_pac_ids = set()
        attributes_for_pac = self._get_attributes_for_pac_id(pac_url=r.pac_id, 
                                                            restrict_to_attribute_groups = r.restrict_to_attribute_groups)
        attributes_for_pac_id.append(attributes_for_pac)
        ref = self._get_referenced_pac_ids(attributes_for_pac)
        if ref:
            referenced_pac_ids.update(ref)
            
        # also find attributes of referenced pac-ids 
        if r.do_forward_lookup:
            for pac_url in referenced_pac_ids:
                attributes_for_pac = self._get_attributes_for_pac_id(pac_url=pac_url, 
                                                                    restrict_to_attribute_groups = r.restrict_to_attribute_groups)
                attributes_for_pac_id.append(attributes_for_pac)

        # add translations
        response_language = self._find_response_language(r.language_preferences)
        for e in attributes_for_pac_id:
            self._add_display_names(e, response_language)
            
        response = AttributeResponsePayload(data=attributes_for_pac_id, language=response_language
                    ).to_json()
        return response
    



    def _get_attributes_for_pac_id(self, pac_url:str, restrict_to_attribute_groups:list[str]|None=None ) -> AttributesOfPACID:
        attribute_groups = []
        if restrict_to_attribute_groups:
            relevant_data_sources = [ds for ds in self._attribute_group_data_sources if ds.attribute_group_key in restrict_to_attribute_groups]
        else:
            relevant_data_sources = self._attribute_group_data_sources
        for ds in relevant_data_sources:
            try: 
                ag = ds.attributes(pac_url)
                if ag:
                    attribute_groups.append(ag)
            except Exception as e:
                e.add_note(f'Attribute Source {ds.attribute_group_key} encountered an error')
                traceback.print_exc()
                raise e
                        
        return AttributesOfPACID(pac_id=pac_url, # return the pac_url as given, i.e. with the extension if there was one
                                 attribute_groups=attribute_groups)
        
    

    def _get_referenced_pac_ids(self, attributes_for_pac:AttributesOfPACID):
        referenced_pacs = []
        for ag in attributes_for_pac.attribute_groups:
            for a in ag.attributes.values() :
                for e in a.items:
                    if isinstance(e, ReferenceAttributeItemsElement):
                        try:
                            PAC_ID.from_url(e.value)
                            referenced_pacs.append(e.value)
                        except Exception:
                            pass
        return referenced_pacs
            
    
    def _add_display_names(self, attributes_of_pac:AttributesOfPACID, language:str) -> str:
        ''' 
        adds the display names in the requested language to attribute group and attributes.
        if no translation can be found in this language it IMMEDIATELY falls back to some - probably inappropriate- magic.
        Note: The server checks for completeness of translations at initialization. Make sure to resolve warnings there and 
        this function should never get into the situation not to find translations.
        '''
        for ag in attributes_of_pac.attribute_groups:
            if dn := self._get_display_name_for_key(ag.group_key, language):
                ag.group_label = dn
            else:
                ag.group_label = self.fallback_label(ag.group_key)
                rich.print(f"[yellow]WARNING:[/yellow] No translation for '{ag.group_key}' in '{language}'. Falling back to '{ag.group_label}'")
            for a in ag.attributes.values():
                if dn := self._get_display_name_for_key(a.key, language):
                    a.label = dn
                elif a.label:
                    pass # label was provided. let's use it as is
                else:
                    a.label = self.fallback_label(a.key)
                    rich.print(f"[yellow]WARNING:[/yellow] No translation for '{a.key}' in '{language}'. Falling back to '{a.label}' ")
                    
        
    def fallback_label(self, key:str):
        l = key.split('/')[-1]
        l = re.sub(r'([a-z])([A-Z])', r'\1 \2', l)
        l = re.sub(r'[-_]', ' ', l)
        l = string.capwords(l)
        return l                 
            
        
    def _get_display_name_for_key(self, key, language:str): 
        '''call this only with a language you know there is a translation for'''
        for tds in self._translation_data_sources:
            if term := tds.get_translations_for(key):
                return term.in_language(language)    
        warnings.warn(f'No translation for {key}.')
        return None
                
                
    def _find_response_language(self, requested_languages:LanguageAccept):
        '''finds the language the server will respond in'''
        if not requested_languages:
            return self._default_language
        best_match = requested_languages.best_match(self._supported_languages, default="en")
        return best_match

                
                
        
    def capabilities(self) -> ServerCapabilities:
        return ServerCapabilities(supported_languages=self._supported_languages,
                                  default_language=self._default_language,
                                  available_attribute_groups= [ds.attribute_group_key for ds in self._attribute_group_data_sources])
    

            








    
        
        
        
    

    
    
    